from __future__ import print_function

import random
import unittest

from EmbCompress import expand, compress


class TestReadHus(unittest.TestCase):

    def test_fake_compression(self):
        for i in range(10):
            s = random.randint(10, 1000)
            test_bytes = bytearray(random.getrandbits(8) for _ in range(s))
            compressed_bytes = compress(test_bytes)
            uncompressed = bytearray(expand(compressed_bytes, len(test_bytes)))
            self.assertEqual(test_bytes, uncompressed)



from __future__ import print_function

import unittest

from pattern_for_tests import *


class TestJson(unittest.TestCase):

    def test_metadata_write_read_json(self):
        file1 = "metadata.json"
        pattern = EmbPattern()
        for i in range(4, 20):
            pattern.add_thread(EmbThread("random"))
        pattern.extras['name'] = "My Embroidery."
        pattern.extras['bytes'] = b'\00these are bytes'
        pattern.extras['bytes2'] = bytearray(b'this is bytearray')
        pattern.extras['value'] = 208
        write(pattern, file1)
        w_pattern = read(file1)
        self.assertEqual(w_pattern.extras['name'], "My Embroidery.")
        self.assertEqual(w_pattern.extras['value'], 208)
        self.assertEqual(len(pattern.threadlist), len(w_pattern.threadlist))
        for q in range(0, len(pattern.threadlist)):
            self.assertEqual(pattern.threadlist[q], w_pattern.threadlist[q])

        self.addCleanup(os.remove, file1)


    def test_colors_write_read_json(self):
        file1 = "color.json"
        for m in range(0, 50):
            pattern = EmbPattern()
            for i in range(4, 20):
                pattern.add_thread(EmbThread("random"))
            write(pattern, file1)
            w_pattern = read(file1)
            self.assertEqual(len(pattern.threadlist), len(w_pattern.threadlist))
            for q in range(0, len(pattern.threadlist)):
                self.assertEqual(pattern.threadlist[q], w_pattern.threadlist[q])
        self.addCleanup(os.remove, file1)

    def test_write_read_json(self):
        file1 = "test.json"
        pattern = get_shift_pattern()
        write(pattern, file1)
        w_pattern = read(file1)
        self.assertEqual(len(pattern.threadlist), len(w_pattern.threadlist))
        for q in range(0, len(pattern.threadlist)):
            self.assertEqual(pattern.threadlist[q], w_pattern.threadlist[q])
        for q in range(0, len(pattern.stitches)):
            self.assertEqual(pattern.stitches[q], w_pattern.stitches[q])
        self.addCleanup(os.remove, file1)



def parse(f):
    comment = None
    code = ""
    value = ""
    command_map = {}
    ord_a = ord('a')
    ord_A = ord('A')
    ord_z = ord('z')
    ord_Z = ord('Z')
    while True:
        byte = f.read(1)
        if byte is None:
            break
        if len(byte) == 0:
            break
        is_end = byte == b'\n' or byte == b'\r'
        if comment is not None:
            if byte == b')' or is_end:
                command_map['comment'] = comment
                comment = None
                if not is_end:
                    continue
            else:
                try:
                    comment += byte.decode('utf8')
                except UnicodeDecodeError:
                    pass  # skip utf8 fail
                continue
        if byte == b'(':
            comment = ""
            continue
        elif byte == b';':
            comment = ""
            continue
        elif byte == b'\t':
            continue
        elif byte == b' ':
            continue
        elif byte == b'/' and len(code) == 0:
            continue
        b = ord(byte)
        if ord('0') <= b <= ord('9') \
                or byte == b'+' \
                or byte == b'-' \
                or byte == b'.':
            value += byte.decode('utf8')
            continue

        if ord_A <= b <= ord_Z:
            b = b - ord_A + ord_a
            byte = chr(b)
        is_letter = ord_a <= b <= ord_z
        if (is_letter or is_end) and len(code) != 0:
            command_map[code] = float(value)
            code = ""
            value = ""
        if is_letter:
            code += str(byte)
            continue
        elif is_end:
            if len(command_map) == 0:
                continue
            yield command_map
            command_map = {}
            code = ""
            value = ""
            continue


def read(f, out, settings=None):
    absolute_mode = True
    flip_x = -1  # Assumes the GCode is flip_x, -1 is flip, 1 is normal
    flip_y = -1  # Assumes the Gcode is flip_y,  -1 is flip, 1 is normal
    scale = 10.0  # Initially assume mm mode G20.
    for gc in parse(f):
        if 'comment' in gc:
            comment = gc['comment']
            if 'Thread' in comment:
                split = comment.split(" ")
                out.add_thread(split[1])

        if 'g' in gc:
            if 'x' in gc and 'y' in gc and gc['g'] == 0.0 or gc['g'] == 1.0:
                if absolute_mode:
                    out.stitch_abs(gc['x'] * scale * flip_x, gc['y'] * scale * flip_y)
                else:
                    out.stitch(gc['x'] * scale * flip_x, gc['y'] * scale * flip_y)
                continue
            if gc['g'] == 21.0 or gc['g'] == 71.0:
                scale = 10.0  # g20 is mm mode. 10 1/10th mm in a mm.
            elif gc['g'] == 20.0 or gc['g'] == 70.0:
                scale = 254  # g20 is inch mode. 254 1/10th mm in an inch.
            elif gc['g'] == 90.0:
                absolute_mode = True
            elif gc['g'] == 91.0:
                absolute_mode = False
        if 'm' in gc:
            v = gc['m']
            if v == 30 or v == 2:
                out.end()
            elif v == 0 or v == 1:
                out.color_change()



from __future__ import print_function

import unittest

from pattern_for_tests import *
from pyembroidery import *


class TestWrites(unittest.TestCase):

    def position_equals(self, stitches, j, k):
        self.assertEqual(stitches[j][:1], stitches[k][:1])

    def test_write_png(self):
        file1 = "file.png"
        write_png(get_shift_pattern(), file1, {"background": "#F00", "linewidth": 5})
        self.addCleanup(os.remove, file1)

    def test_write_dst_read_dst(self):
        file1 = "file.dst"
        write_dst(get_big_pattern(), file1)
        dst_pattern = read_dst(file1)
        self.assertEqual(len(dst_pattern.threadlist), 0)
        self.assertEqual(dst_pattern.count_stitch_commands(COLOR_CHANGE), 15)
        self.assertIsNotNone(dst_pattern)
        self.assertEqual(dst_pattern.count_stitch_commands(STITCH), 5 * 16)
        self.position_equals(dst_pattern.stitches, 0, -1)
        print("dst: ", dst_pattern.stitches)
        self.addCleanup(os.remove, file1)

    def test_write_exp_read_exp(self):
        file1 = "file.exp"
        write_exp(get_big_pattern(), file1)
        exp_pattern = read_exp(file1)
        self.assertEqual(len(exp_pattern.threadlist), 0)
        self.assertEqual(exp_pattern.count_stitch_commands(COLOR_CHANGE), 15)
        self.assertIsNotNone(exp_pattern)
        self.assertEqual(exp_pattern.count_stitch_commands(STITCH), 5 * 16)
        self.position_equals(exp_pattern.stitches, 0, -1)
        print("exp: ", exp_pattern.stitches)
        self.addCleanup(os.remove, file1)

    def test_write_vp3_read_vp3(self):
        file1 = "file.vp3"
        write_vp3(get_big_pattern(), file1)
        vp3_pattern = read_vp3(file1)
        self.assertEqual(len(vp3_pattern.threadlist), vp3_pattern.count_stitch_commands(COLOR_CHANGE) + 1)
        self.assertEqual(vp3_pattern.count_stitch_commands(COLOR_CHANGE), 15)
        self.assertIsNotNone(vp3_pattern)
        self.assertEqual(vp3_pattern.count_stitch_commands(STITCH), 5 * 16)
        self.position_equals(vp3_pattern.stitches, 0, -1)
        print("vp3: ", vp3_pattern.stitches)
        self.addCleanup(os.remove, file1)

    def test_write_jef_read_jef(self):
        file1 = "file.jef"
        write_jef(get_big_pattern(), file1)
        jef_pattern = read_jef(file1)
        self.assertEqual(len(jef_pattern.threadlist), jef_pattern.count_stitch_commands(COLOR_CHANGE) + 1)
        self.assertEqual(jef_pattern.count_stitch_commands(COLOR_CHANGE), 15)
        self.assertIsNotNone(jef_pattern)
        self.assertEqual(jef_pattern.count_stitch_commands(STITCH), 5 * 16)
        self.position_equals(jef_pattern.stitches, 0, -1)
        print("jef: ", jef_pattern.stitches)
        self.addCleanup(os.remove, file1)

    def test_write_pec_read_pec(self):
        file1 = "file.pec"
        write_pec(get_big_pattern(), file1)
        pec_pattern = read_pec(file1)
        self.assertEqual(len(pec_pattern.threadlist), pec_pattern.count_stitch_commands(COLOR_CHANGE) + 1)
        self.assertEqual(pec_pattern.count_stitch_commands(COLOR_CHANGE), 15)
        self.assertIsNotNone(pec_pattern)
        self.assertEqual(pec_pattern.count_stitch_commands(STITCH), 5 * 16)
        self.position_equals(pec_pattern.stitches, 0, -1)
        print("pec: ", pec_pattern.stitches)
        self.addCleanup(os.remove, file1)

    def test_write_pes_read_pes(self):
        file1 = "file.pes"
        write_pes(get_big_pattern(), file1)
        pes_pattern = read_pes(file1)
        self.assertEqual(len(pes_pattern.threadlist), pes_pattern.count_stitch_commands(COLOR_CHANGE) + 1)
        self.assertEqual(pes_pattern.count_stitch_commands(COLOR_CHANGE), 15)
        self.assertIsNotNone(pes_pattern)
        self.assertEqual(pes_pattern.count_stitch_commands(STITCH), 5 * 16)
        self.position_equals(pes_pattern.stitches, 0, -1)
        print("pes: ", pes_pattern.stitches)
        self.addCleanup(os.remove, file1)

    def test_write_xxx_read_xxx(self):
        file1 = "file.xxx"
        write_xxx(get_big_pattern(), file1)
        pattern = read_xxx(file1)
        self.assertEqual(len(pattern.threadlist), pattern.count_stitch_commands(COLOR_CHANGE) + 1)
        self.assertEqual(pattern.count_stitch_commands(COLOR_CHANGE), 15)
        self.assertIsNotNone(pattern)
        self.assertEqual(pattern.count_stitch_commands(STITCH), 5 * 16)
        self.position_equals(pattern.stitches, 0, -1)
        print("xxx: ", pattern.stitches)
        self.addCleanup(os.remove, file1)

    def test_write_u01_read_u01(self):
        file1 = "file.u01"
        write_u01(get_big_pattern(), file1)
        u01_pattern = read_u01(file1)
        self.assertEqual(len(u01_pattern.threadlist), 0)
        self.assertEqual(u01_pattern.count_stitch_commands(NEEDLE_SET), 16)
        self.assertIsNotNone(u01_pattern)
        self.assertEqual(u01_pattern.count_stitch_commands(STITCH), 5 * 16)
        self.position_equals(u01_pattern.stitches, 0, -1)
        print("u01: ", u01_pattern.stitches)
        self.addCleanup(os.remove, file1)

    def test_write_csv_read_csv(self):
        file1 = "file.csv"
        write_csv(get_big_pattern(), file1, {"encode": True})
        csv_pattern = read_csv(file1)
        self.assertIsNotNone(csv_pattern)
        self.assertEqual(len(csv_pattern.threadlist), csv_pattern.count_stitch_commands(COLOR_CHANGE) + 1)
        self.assertEqual(csv_pattern.count_stitch_commands(COLOR_CHANGE), 15)
        self.assertEqual(csv_pattern.count_stitch_commands(STITCH), 5 * 16)
        self.position_equals(csv_pattern.stitches, 0, -1)
        print("csv: ", csv_pattern.stitches)
        self.addCleanup(os.remove, file1)

    def test_write_gcode_read_gcode(self):
        file1 = "file.gcode"
        write_gcode(get_big_pattern(), file1)
        gcode_pattern = read_gcode(file1)
        self.assertIsNotNone(gcode_pattern)
        thread_count = len(gcode_pattern.threadlist)
        change_count = gcode_pattern.count_stitch_commands(COLOR_CHANGE) + 1
        print(thread_count)
        print(change_count)

        self.assertEqual(thread_count, change_count)
        self.assertEqual(gcode_pattern.count_stitch_commands(COLOR_CHANGE), 15)
        self.assertEqual(gcode_pattern.count_stitch_commands(STITCH), 5 * 16)
        self.position_equals(gcode_pattern.stitches, 0, -1)
        self.addCleanup(os.remove, file1)

    def test_write_txt(self):
        file1 = "file.txt"
        write_txt(get_big_pattern(), file1)
        write_txt(get_big_pattern(), file1, {"mimic": True})
        self.addCleanup(os.remove, file1)

    def test_write_pes_mismatched(self):
        file1 = "file.pes"
        pattern = EmbPattern()
        pattern += "red"
        pattern += "red"
        pattern += "red"
        pattern += (0, 0), (100, 100)
        pattern += COLOR_CHANGE
        pattern += (100, 0), (0, 100)
        pattern += COLOR_CHANGE
        pattern += (0, 0), (100, 100)
        pattern += COLOR_CHANGE
        pattern += (100, 0), (0, 100)
        write_pes(pattern, file1, {"version": "6t"})
        write_pes(pattern, file1)
        self.addCleanup(os.remove, file1)

    def test_pes_writes_stop(self):
        """Test if pes can read/write a stop command."""
        file1 = "file.pes"
        pattern = EmbPattern()
        pattern += "red"
        pattern += (0, 0), (100, 100)
        pattern += STOP
        pattern += (100, 0), (0, 100)
        pattern += "blue"
        pattern += (0, 0), (100, 100)
        pattern += STOP
        pattern += (100, 0), (0, 100)
        write_pes(pattern, file1, {"version": "6t"})
        loaded = read_pes(file1)
        self.assertEqual(pattern.count_stitch_commands(STOP), 2)
        self.assertEqual(pattern.count_stitch_commands(COLOR_CHANGE), 1)
        self.assertEqual(pattern.count_threads(), 2)
        self.assertEqual(loaded.count_stitch_commands(STOP), 2)
        self.assertEqual(loaded.count_stitch_commands(COLOR_CHANGE), 1)
        self.assertEqual(loaded.count_threads(), 2)
        self.addCleanup(os.remove, file1)




import os.path

from .EmbPattern import EmbPattern

supported_formats = EmbPattern.supported_formats
convert = EmbPattern.convert
get_extension_by_filename = EmbPattern.get_extension_by_filename

read_embroidery = EmbPattern.read_embroidery
read_dst = EmbPattern.read_dst
read_pec = EmbPattern.read_pec
read_pes = EmbPattern.read_pes
read_exp = EmbPattern.read_exp
read_vp3 = EmbPattern.read_vp3
read_jef = EmbPattern.read_jef
read_u01 = EmbPattern.read_u01
read_csv = EmbPattern.read_csv
read_gcode = EmbPattern.read_gcode
read_xxx = EmbPattern.read_xxx
read = EmbPattern.static_read

write_embroidery = EmbPattern.write_embroidery
write_dst = EmbPattern.write_dst
write_pec = EmbPattern.write_pec
write_pes = EmbPattern.write_pes
write_exp = EmbPattern.write_exp
write_vp3 = EmbPattern.write_vp3
write_jef = EmbPattern.write_jef
write_u01 = EmbPattern.write_u01
write_csv = EmbPattern.write_csv
write_txt = EmbPattern.write_txt
write_gcode = EmbPattern.write_gcode
write_xxx = EmbPattern.write_xxx
write_svg = EmbPattern.write_svg
write_png = EmbPattern.write_png
write = EmbPattern.static_write
is_str = EmbPattern.is_str


import os

import pyembroidery.A100Reader as A100Reader
import pyembroidery.A10oReader as A10oReader
# import pyembroidery.ArtReader as ArtReader
import pyembroidery.BroReader as BroReader
import pyembroidery.ColReader as ColReader
import pyembroidery.ColWriter as ColWriter
import pyembroidery.CsvReader as CsvReader
import pyembroidery.CsvWriter as CsvWriter
import pyembroidery.DatReader as DatReader
import pyembroidery.DsbReader as DsbReader
import pyembroidery.DstReader as DstReader
import pyembroidery.DstWriter as DstWriter
import pyembroidery.DszReader as DszReader
import pyembroidery.EdrReader as EdrReader
import pyembroidery.EdrWriter as EdrWriter
import pyembroidery.EmdReader as EmdReader
import pyembroidery.ExpReader as ExpReader
import pyembroidery.ExpWriter as ExpWriter
import pyembroidery.ExyReader as ExyReader
import pyembroidery.FxyReader as FxyReader
import pyembroidery.GcodeReader as GcodeReader
import pyembroidery.GcodeWriter as GcodeWriter
import pyembroidery.GtReader as GtReader
import pyembroidery.HusReader as HusReader
import pyembroidery.InbReader as InbReader
import pyembroidery.InfReader as InfReader
import pyembroidery.InfWriter as InfWriter
import pyembroidery.JefReader as JefReader
import pyembroidery.JefWriter as JefWriter
import pyembroidery.JpxReader as JpxReader
import pyembroidery.JsonReader as JsonReader
import pyembroidery.JsonWriter as JsonWriter
import pyembroidery.KsmReader as KsmReader
import pyembroidery.MaxReader as MaxReader
import pyembroidery.MitReader as MitReader
import pyembroidery.NewReader as NewReader
import pyembroidery.PcdReader as PcdReader
import pyembroidery.PcmReader as PcmReader
import pyembroidery.PcqReader as PcqReader
import pyembroidery.PcsReader as PcsReader
import pyembroidery.PecReader as PecReader
import pyembroidery.PecWriter as PecWriter
import pyembroidery.PesReader as PesReader
import pyembroidery.PesWriter as PesWriter
import pyembroidery.PhbReader as PhbReader
import pyembroidery.PhcReader as PhcReader
import pyembroidery.PmvReader as PmvReader
import pyembroidery.PmvWriter as PmvWriter
import pyembroidery.PngWriter as PngWriter
import pyembroidery.SewReader as SewReader
import pyembroidery.ShvReader as ShvReader
import pyembroidery.SpxReader as SpxReader
import pyembroidery.StcReader as StcReader
import pyembroidery.StxReader as StxReader
import pyembroidery.SvgWriter as SvgWriter
import pyembroidery.TapReader as TapReader
import pyembroidery.TbfReader as TbfReader
import pyembroidery.TxtWriter as TxtWriter
import pyembroidery.U01Reader as U01Reader
import pyembroidery.U01Writer as U01Writer
import pyembroidery.Vp3Reader as Vp3Reader
import pyembroidery.Vp3Writer as Vp3Writer
import pyembroidery.XxxReader as XxxReader
import pyembroidery.XxxWriter as XxxWriter
# import pyembroidery.ZhsReader as ZhsReader
import pyembroidery.ZxyReader as ZxyReader
from .EmbEncoder import Transcoder as Normalizer
from .EmbFunctions import *
from .EmbThread import EmbThread


class EmbPattern:
    def __init__(self, *args, **kwargs):
        self.stitches = []  # type: list
        self.threadlist = []  # type: list
        self.extras = {}  # type: dict
        # filename, name, category, author, keywords, comments, are typical
        self._previousX = 0  # type: float
        self._previousY = 0  # type: float
        len_args = len(args)
        if len_args >= 1:
            arg0 = args[0]
            if isinstance(arg0, EmbPattern):
                self.stitches = arg0.stitches[:]
                self.threadlist = arg0.threadlist[:]
                self.extras.update(arg0.extras)
                self._previousX = arg0._previousX
                self._previousY = arg0._previousY
                return
            if len(args) >= 2:
                settings = args[1]
            elif 'settings' in kwargs:
                settings = kwargs['settings']
            else:
                settings = kwargs
            if isinstance(arg0, str):
                EmbPattern.static_read(arg0, settings=settings, pattern=self)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __eq__(self, other):
        if not isinstance(other, EmbPattern):
            return False
        if self.stitches != other.stitches:
            return False
        if self.threadlist != other.threadlist:
            return False
        if self.extras != other.extras:
            return False
        return True

    def __str__(self):
        if "name" in self.extras:
            return "EmbPattern %s (commands: %3d, threads: %3d)" % \
                   (self.extras["name"], len(self.stitches), len(self.threadlist))
        return "EmbPattern (commands: %3d, threads: %3d)" % (len(self.stitches), len(self.threadlist))

    def __len__(self):
        return len(self.stitches)

    def __getitem__(self, item):
        if isinstance(item, str):
            return self.extras[item]
        return self.stitches[item]

    def __setitem__(self, key, value):
        if isinstance(key, str):
            self.extras[key] = value
        else:
            self.stitches[key] = value

    def __copy__(self):
        return self.copy()

    def __deepcopy__(self):
        return self.copy()

    def __iadd__(self, other):
        if isinstance(other, EmbPattern):
            self.add_pattern(other)
        elif isinstance(other, EmbThread) or isinstance(other, str):
            self.add_thread(other)
            for i in range(0, len(self.stitches)):
                data = self.stitches[i][2] & COMMAND_MASK
                if data == STITCH or data == SEW_TO or data == NEEDLE_AT:
                    self.color_change()
                    break  # Only add color change if stitching exists.
        elif isinstance(other, int):
            self.add_command(other)
        elif isinstance(other, list) or isinstance(other, tuple):  # tuple or list
            if len(other) == 0:
                return
            v = other[0]
            if isinstance(v, list) or isinstance(v, tuple):  # tuple or list of tuple or lists
                for v in other:
                    x = v[0]
                    y = v[1]
                    try:
                        cmd = v[2]
                    except IndexError:
                        cmd = STITCH
                    self.add_stitch_absolute(cmd, x, y)
            elif isinstance(v, complex):  # tuple or list of complex
                for v in other:
                    x = v.real
                    y = v.imag
                    self.add_stitch_absolute(STITCH, x, y)
            elif isinstance(v, int) or isinstance(v, float):  # tuple or list of numbers.
                i = 0
                ie = len(other)
                while i < ie:
                    self.add_stitch_absolute(STITCH, other[i], other[i + 1])
                    i += 2
            elif isinstance(v, str):
                self.extras[v] = other[1]
        else:
            raise ValueError()
        return self

    def __add__(self, other):
        p = self.copy()
        p.add_pattern(other)
        return p

    def __radd__(self, other):
        p = other.copy()
        p.add_pattern(self)
        return p

    def copy(self):
        emb_pattern = EmbPattern()
        emb_pattern.stitches = self.stitches[:]
        emb_pattern.threadlist = self.threadlist[:]
        emb_pattern.extras.update(self.extras)
        emb_pattern._previousX = self._previousX
        emb_pattern._previousY = self._previousY
        return emb_pattern

    def clear(self):
        self.stitches = []
        self.threadlist = []
        self.extras = {}
        self._previousX = 0
        self._previousY = 0

    def read(self, filename, **settings):
        EmbPattern.static_read(filename, settings=settings, pattern=self)

    def write(self, filename, **settings):
        EmbPattern.static_write(self, filename, settings=settings)

    def move(self, dx=0, dy=0, position=None):
        """Move dx, dy"""
        if position is None:
            self.add_stitch_relative(JUMP, dx, dy)
        else:
            self.insert_stitch_relative(position, JUMP, dx, dy)

    def move_abs(self, x, y, position=None):
        """Move absolute x, y"""
        if position is None:
            self.add_stitch_absolute(JUMP, x, y)
        else:
            self.insert(position, JUMP, x, y)

    def stitch(self, dx=0, dy=0, position=None):
        """Stitch dx, dy"""
        if position is None:
            self.add_stitch_relative(STITCH, dx, dy)
        else:
            self.insert_stitch_relative(position, STITCH, dx, dy)

    def stitch_abs(self, x, y, position=None):
        """Stitch absolute x, y"""
        if position is None:
            self.add_stitch_absolute(STITCH, x, y)
        else:
            self.insert(position, STITCH, x, y)

    def stop(self, dx=0, dy=0, position=None):
        """Stop dx, dy"""
        if position is None:
            self.add_stitch_relative(STOP, dx, dy)
        else:
            self.insert_stitch_relative(position, STOP, dx, dy)

    def trim(self, dx=0, dy=0, position=None):
        """Trim dx, dy"""
        if position is None:
            self.add_stitch_relative(TRIM, dx, dy)
        else:
            self.insert_stitch_relative(position, TRIM, dx, dy)

    def color_change(self, dx=0, dy=0, position=None):
        """Color Change dx, dy"""
        if position is None:
            self.add_stitch_relative(COLOR_CHANGE, dx, dy)
        else:
            self.insert_stitch_relative(position, COLOR_CHANGE, dx, dy)

    def needle_change(self, needle=0, dx=0, dy=0, position=None):
        """Needle change, needle, dx, dy"""
        cmd = encode_thread_change(NEEDLE_SET, None, needle)
        if position is None:
            self.add_stitch_relative(cmd, dx, dy)
        else:
            self.insert_stitch_relative(position, cmd, dx, dy)

    def sequin_eject(self, dx=0, dy=0, position=None):
        """Eject Sequin dx, dy"""
        if position is None:
            self.add_stitch_relative(SEQUIN_EJECT, dx, dy)
        else:
            self.insert_stitch_relative(position, SEQUIN_EJECT, dx, dy)

    def sequin_mode(self, dx=0, dy=0, position=None):
        """Eject Sequin dx, dy"""
        if position is None:
            self.add_stitch_relative(SEQUIN_MODE, dx, dy)
        else:
            self.insert_stitch_relative(position, SEQUIN_MODE, dx, dy)

    def end(self, dx=0, dy=0, position=None):
        """End Design dx, dy"""
        if position is None:
            self.add_stitch_relative(END, dx, dy)
        else:
            self.insert_stitch_relative(position, END, dx, dy)

    def add_thread(self, thread):
        """Adds thread to design.
        Note: this has no effect on stitching and can be done at any point."""
        if isinstance(thread, EmbThread):
            self.threadlist.append(thread)
        else:
            thread_object = EmbThread()
            thread_object.set(thread)
            self.threadlist.append(thread_object)

    def metadata(self, name, data):
        """Adds select metadata to design.
        Note: this has no effect on stitching and can be done at any point."""
        self.extras[name] = data

    def get_metadata(self, name, default=None):
        return self.extras.get(name, default)

    def bounds(self):
        """Returns the bounds of the stitch data:
        min_x, min_y, max_x, max_y"""
        min_x = float('inf')
        min_y = float('inf')
        max_x = -float('inf')
        max_y = -float('inf')

        for stitch in self.stitches:
            if stitch[0] > max_x:
                max_x = stitch[0]
            if stitch[0] < min_x:
                min_x = stitch[0]
            if stitch[1] > max_y:
                max_y = stitch[1]
            if stitch[1] < min_y:
                min_y = stitch[1]
        return min_x, min_y, max_x, max_y

    extends = bounds
    extents = bounds

    def count_stitch_commands(self, command):
        count = 0
        for stitch in self.stitches:
            flags = stitch[2] & COMMAND_MASK
            if flags == command:
                count += 1
        return count

    def count_color_changes(self):
        return self.count_stitch_commands(COLOR_CHANGE)

    def count_needle_sets(self):
        return self.count_stitch_commands(NEEDLE_SET)

    def count_stitches(self):
        return len(self.stitches)

    def count_threads(self):
        return len(self.threadlist)

    @staticmethod
    def get_random_thread():
        thread = EmbThread()
        thread.set("random")
        thread.description = "Random"
        return thread

    def get_thread_or_filler(self, index):
        if len(self.threadlist) <= index:
            return self.get_random_thread()
        else:
            return self.threadlist[index]

    def get_thread(self, index):
        return self.threadlist[index]

    def get_match_commands(self, command):
        for stitch in self.stitches:
            flags = stitch[2] & COMMAND_MASK
            if flags == command:
                yield stitch

    def get_as_stitchblock(self):
        stitchblock = []
        thread = self.get_thread_or_filler(0)
        thread_index = 1
        for stitch in self.stitches:
            flags = stitch[2] & COMMAND_MASK
            if flags == STITCH:
                stitchblock.append(stitch)
            else:
                if len(stitchblock) > 0:
                    yield (stitchblock, thread)
                    stitchblock = []
                if flags == COLOR_CHANGE:
                    thread = self.get_thread_or_filler(thread_index)
                    thread_index += 1
        if len(stitchblock) > 0:
            yield (stitchblock, thread)

    def get_as_command_blocks(self):
        last_pos = 0
        last_command = NO_COMMAND
        for pos, stitch in enumerate(self.stitches):
            command = stitch[2] & COMMAND_MASK
            if command == last_command or last_command == NO_COMMAND:
                last_command = command
                continue
            last_command = command
            yield self.stitches[last_pos:pos]
            last_pos = pos
        yield self.stitches[last_pos:]

    def get_as_colorblocks(self):
        """
        Returns a generator for colorblocks. Color blocks defined with color_breaks will have
        the command omitted whereas color blocks delimited with color_change will end with the
        color_change command, and if delimited with needle_set, the blocks will begin the new
        color block with the needle_set.
        """
        thread_index = 0
        colorblock_start = 0

        for pos, stitch in enumerate(self.stitches):
            command = stitch[2] & COMMAND_MASK
            if command == COLOR_BREAK:
                if colorblock_start != pos:
                    thread = self.get_thread_or_filler(thread_index)
                    thread_index += 1
                    yield self.stitches[colorblock_start:pos], thread
                colorblock_start = pos + 1
                continue
            if command == COLOR_CHANGE:
                thread = self.get_thread_or_filler(thread_index)
                thread_index += 1
                yield self.stitches[colorblock_start:pos + 1], thread
                colorblock_start = pos + 1
                continue
            if command == NEEDLE_SET and colorblock_start != pos:
                thread = self.get_thread_or_filler(thread_index)
                thread_index += 1
                yield self.stitches[colorblock_start:pos], thread
                colorblock_start = pos
                continue

        if colorblock_start != len(self.stitches):
            thread = self.get_thread_or_filler(thread_index)
            yield self.stitches[colorblock_start:], thread

    def get_as_stitches(self):
        """pos, x, y, command, v1, v2, v3"""
        for pos, stitch in enumerate(self.stitches):
            decode = decode_embroidery_command(stitch[2])
            command = decode[0]
            thread = decode[1]
            needle = decode[2]
            order = decode[3]
            yield pos, stitch[0], stitch[1], command, thread, needle, order

    def get_unique_threadlist(self):
        return set(self.threadlist)

    def get_singleton_threadlist(self):
        singleton = []
        last_thread = None
        for thread in self.threadlist:
            if thread != last_thread:
                singleton.append(thread)
            last_thread = thread
        return singleton

    def move_center_to_origin(self):
        extends = self.bounds()
        cx = round((extends[2] - extends[0]) / 2.0)
        cy = round((extends[3] - extends[1]) / 2.0)
        self.translate(-cx, -cy)

    def translate(self, dx, dy):
        for stitch in self.stitches:
            stitch[0] += dx
            stitch[1] += dy

    def transform(self, matrix):
        for stitch in self.stitches:
            matrix.apply(stitch)

    def fix_color_count(self):
        """Ensure that there are threads for all color blocks."""
        thread_index = 0
        init_color = True
        for stitch in self.stitches:
            data = stitch[2] & COMMAND_MASK
            if data == STITCH or data == SEW_TO or data == NEEDLE_AT:
                if init_color:
                    thread_index += 1
                    init_color = False
            elif data == COLOR_CHANGE or data == COLOR_BREAK or data == NEEDLE_SET:
                init_color = True
        while len(self.threadlist) < thread_index:
            self.add_thread(self.get_thread_or_filler(len(self.threadlist)))

    def add_stitch_absolute(self, cmd, x=0, y=0):
        """Add a command at the absolute location: x, y"""
        self.stitches.append([x, y, cmd])
        self._previousX = x
        self._previousY = y

    def add_stitch_relative(self, cmd, dx=0, dy=0):
        """Add a command relative to the previous location"""
        x = self._previousX + dx
        y = self._previousY + dy
        self.add_stitch_absolute(cmd, x, y)

    def insert_stitch_relative(self, position, cmd, dx=0, dy=0):
        """Insert a relative stitch into the pattern. The stitch is relative to the stitch before it.
        If inserting at position 0, it's relative to 0,0. If appending, add is called, updating the positioning.
        """
        if position < 0:
            position += len(self.stitches)  # I need positive positions.
        if position == 0:
            self.stitches.insert(0, [dx, dy, TRIM])  # started (0,0)
        elif position == len(self.stitches) or position is None:  # This is properly just an add.
            self.add_stitch_relative(cmd, dx, dy)
        elif 0 < position < len(self.stitches):
            p = self.stitches[position - 1]
            x = p[0] + dx
            y = p[1] + dy
            self.stitches.insert(position, [x, y, TRIM])

    def insert(self, position, cmd, x=0, y=0):
        """Insert a stitch or command"""
        self.stitches.insert(position, [x, y, cmd])

    def prepend_command(self, cmd, x=0, y=0):
        """Prepend a command, without treating parameters as locations"""
        self.stitches.insert(0, [x, y, cmd])

    def add_command(self, cmd, x=0, y=0):
        """Add a command, without treating parameters as locations
         that require an update"""
        self.stitches.append([x, y, cmd])

    def add_block(self, block, thread=None):
        if thread is not None:
            self.add_thread(thread)
        if block is None:
            return

        if isinstance(block, list) or isinstance(block, tuple):
            if len(block) == 0:
                return
            v = block[0]
            if isinstance(v, list) or isinstance(v, tuple):
                for v in block:
                    x = v[0]
                    y = v[1]
                    try:
                        cmd = v[2]
                    except IndexError:
                        cmd = STITCH
                    self.add_stitch_absolute(cmd, x, y)
            elif isinstance(v, complex):
                for v in block:
                    x = v.real
                    y = v.imag
                    self.add_stitch_absolute(STITCH, x, y)
            elif isinstance(v, int) or isinstance(v, float):
                i = 0
                ie = len(block)
                while i < ie:
                    self.add_stitch_absolute(STITCH, block[i], block[i + 1])
                    i += 2
        self.add_command(COLOR_BREAK)

    def add_stitchblock(self, stitchblock):
        threadlist = self.threadlist
        block = stitchblock[0]
        thread = stitchblock[1]
        if len(threadlist) == 0 or thread is not threadlist[-1]:
            threadlist.append(thread)
            self.add_stitch_relative(COLOR_BREAK)
        else:
            self.add_stitch_relative(SEQUENCE_BREAK)

        for stitch in block:
            try:
                self.add_stitch_absolute(stitch.command, stitch.x, stitch.y)
            except AttributeError:
                self.add_stitch_absolute(stitch[2], stitch[0], stitch[1])

    def add_pattern(self, pattern, dx=None, dy=None, sx=None, sy=None, rotate=None):
        """
        add_pattern merges the given pattern with the current pattern. It accounts for some edge conditions but
        not all of them.

        If there is an end command on the current pattern, that is removed.
        If the color ending the current pattern is equal to the color starting the next those color blocks are merged.
        Any prepended thread change command to the merging pattern is suppressed.

        :param pattern: pattern to add to current pattern
        :return:
        """
        if isinstance(pattern, str):
            pattern = EmbPattern(pattern)
        if self.stitches[-1][2] == END:
            self.stitches = self.stitches[:-1]  # Remove END, if exists
        if dx is not None or dy is not None:
            if dx is None:
                dx = 0
            if dy is None:
                dy = 0
            self.add_command(MATRIX_TRANSLATE, dx, dy)
        if sx is not None or sx is not None:
            if sx is None:
                sx = sy
            if sy is None:
                sy = sx
            self.add_command(MATRIX_SCALE, sx, sy)
        if rotate is not None:
            self.add_command(MATRIX_ROTATE, rotate)
        # Add the new thread only if it's different from the last one
        self.fix_color_count()

        if len(pattern.threadlist) > 0:
            if pattern.threadlist[0] == self.threadlist[-1]:
                self.threadlist.extend(pattern.threadlist[1:])
            else:
                self.threadlist.extend(pattern.threadlist)
                self.color_change()
        join_position = len(self.stitches)
        self.stitches.extend(pattern.stitches)

        for i in range(join_position, len(self.stitches)):
            data = self.stitches[i][2] & COMMAND_MASK
            if data == STITCH or data == SEW_TO or data == NEEDLE_AT:
                break
            elif data == COLOR_CHANGE or data == COLOR_BREAK or data == NEEDLE_SET:
                self.stitches[i][2] = NO_COMMAND
        self.extras.update(pattern.extras)

    def interpolate_duplicate_color_as_stop(self):
        """Processes a pattern replacing any duplicate colors in the threadlist as a stop."""
        thread_index = 0
        init_color = True
        last_change = None
        for position, stitch in enumerate(self.stitches):
            data = stitch[2] & COMMAND_MASK
            if data == STITCH or data == SEW_TO or data == NEEDLE_AT:
                if init_color:
                    try:
                        if last_change is not None and thread_index != 0 and \
                                self.threadlist[thread_index - 1] == self.threadlist[thread_index]:
                            del self.threadlist[thread_index]
                            self.stitches[last_change][2] = STOP
                        else:
                            thread_index += 1
                    except IndexError:  # Non-existant threads cannot double
                        return
                    init_color = False
            elif data == COLOR_CHANGE or data == COLOR_BREAK or data == NEEDLE_SET:
                init_color = True
                last_change = position

    def interpolate_stop_as_duplicate_color(self, thread_change_command=COLOR_CHANGE):
        """Processes a pattern replacing any stop as a duplicate color, and color_change
         or another specified thread_change_command"""
        thread_index = 0
        for position, stitch in enumerate(self.stitches):
            data = stitch[2] & COMMAND_MASK
            if data == STITCH or data == SEW_TO or data == NEEDLE_AT:
                continue
            elif data == COLOR_CHANGE or data == COLOR_BREAK or data == NEEDLE_SET:
                thread_index += 1
            elif data == STOP:
                try:
                    self.threadlist.insert(thread_index, self.threadlist[thread_index])
                    self.stitches[position][2] = thread_change_command
                    thread_index += 1
                except IndexError:  # There are no colors to duplicate
                    return

    def interpolate_frame_eject(self):
        """Processes a pattern replacing jump-stop-jump/jump-stop-end sequences with FRAME_EJECT."""
        mode = 0
        stop_x = None
        stop_y = None
        sequence_start_position = None
        position = 0
        ie = len(self.stitches)
        while position < ie:
            stitch = self.stitches[position]
            data = stitch[2] & COMMAND_MASK
            if data == STITCH or data == SEW_TO or data == NEEDLE_AT or \
                    data == COLOR_CHANGE or data == COLOR_BREAK or data == NEEDLE_SET:
                if mode == 3:
                    del self.stitches[sequence_start_position:position]
                    position = sequence_start_position
                    self.stitches.insert(position, [stop_x, stop_y, FRAME_EJECT])
                    ie = len(self.stitches)
                mode = 0
            elif data == JUMP:
                if mode == 2:
                    mode = 3
                if mode == 0:
                    sequence_start_position = position
                    mode = 1
            elif data == STOP:
                if mode == 1:
                    mode = 2
                    stop_x = stitch[0]
                    stop_y = stitch[1]
            position += 1
        if mode >= 2:  # Frame_eject at end.
            del self.stitches[sequence_start_position:position]
            position = sequence_start_position
            self.stitches.insert(position, [stop_x, stop_y, FRAME_EJECT])

    def interpolate_trims(self, jumps_to_require_trim=None, distance_to_require_trim=None, clipping=True):
        """Processes a pattern adding trims according to the given criteria."""
        i = -1
        ie = len(self.stitches) - 1

        x = 0
        y = 0
        jump_count = 0
        jump_start = 0
        jump_dx = 0
        jump_dy = 0
        jumping = False
        trimmed = True
        while i < ie:
            i += 1
            stitch = self.stitches[i]
            dx = stitch[0] - x
            dy = stitch[1] - y
            x = stitch[0]
            y = stitch[1]
            command = stitch[2] & COMMAND_MASK
            if command == STITCH or command == SEQUIN_EJECT:
                trimmed = False
                jumping = False
            elif command == COLOR_CHANGE or command == NEEDLE_SET or command == TRIM:
                trimmed = True
                jumping = False
            if command == JUMP:
                if not jumping:
                    jump_dx = 0
                    jump_dy = 0
                    jump_count = 0
                    jump_start = i
                    jumping = True
                jump_count += 1
                jump_dx += dx
                jump_dy += dy
                if not trimmed:
                    if jump_count == jumps_to_require_trim or \
                            distance_to_require_trim is not None and \
                            (
                                    abs(jump_dy) > distance_to_require_trim or \
                                    abs(jump_dx) > distance_to_require_trim
                            ):
                        self.trim(position=jump_start)
                        jump_start += 1  # We inserted a position, start jump has moved.
                        i += 1
                        ie += 1
                        trimmed = True
                if clipping and jump_dx == 0 and jump_dy == 0:  # jump displacement is 0, clip trim command.
                    del self.stitches[jump_start:i + 1]
                    i = jump_start - 1
                    ie = len(self.stitches) - 1

    def get_pattern_interpolate_trim(self, jumps_to_require_trim):
        """Gets a processed pattern with untrimmed jumps merged
        and trims added if merged jumps are beyond the given value.
        The expectation is that it has core commands and not
        middle-level commands"""
        new_pattern = EmbPattern()
        i = -1
        ie = len(self.stitches) - 1
        count = 0
        trimmed = True
        while i < ie:
            i += 1
            stitch = self.stitches[i]
            command = stitch[2] & COMMAND_MASK
            if command == STITCH or command == SEQUIN_EJECT:
                trimmed = False
            elif command == COLOR_CHANGE or command == NEEDLE_SET or command == TRIM:
                trimmed = True
            if trimmed or stitch[2] != JUMP:
                new_pattern.add_stitch_absolute(stitch[2],
                                                stitch[0],
                                                stitch[1])
                continue
            while i < ie and command == JUMP:
                i += 1
                stitch = self.stitches[i]
                command = stitch[2]
                count += 1
            if command != JUMP:
                i -= 1
            stitch = self.stitches[i]
            if count >= jumps_to_require_trim:
                new_pattern.trim()
            count = 0
            new_pattern.add_stitch_absolute(stitch[2],
                                            stitch[0],
                                            stitch[1])
        new_pattern.threadlist.extend(self.threadlist)
        new_pattern.extras.update(self.extras)
        return new_pattern

    def get_pattern_merge_jumps(self):
        """Returns a pattern with all multiple jumps merged."""
        new_pattern = EmbPattern()
        i = -1
        ie = len(self.stitches) - 1
        stitch_break = False
        while i < ie:
            i += 1
            stitch = self.stitches[i]
            command = stitch[2] & COMMAND_MASK
            if command == JUMP:
                if stitch_break:
                    continue
                new_pattern.add_command(STITCH_BREAK)
                stitch_break = True
                continue
            new_pattern.add_stitch_absolute(stitch[2],
                                            stitch[0],
                                            stitch[1])
        new_pattern.threadlist.extend(self.threadlist)
        new_pattern.extras.update(self.extras)
        return new_pattern

    def get_stable_pattern(self):
        """Gets a stabilized version of the pattern."""
        stable_pattern = EmbPattern()
        for stitchblock in self.get_as_stitchblock():
            stable_pattern.add_stitchblock(stitchblock)
        stable_pattern.extras.update(self.extras)
        return stable_pattern

    def get_normalized_pattern(self, encode_settings=None):
        """Encodes pattern typically for saving."""
        normal_pattern = EmbPattern()
        transcoder = Normalizer(encode_settings)
        transcoder.transcode(self, normal_pattern)
        return normal_pattern

    def append_translation(self, x, y):
        """Appends translation to the pattern.
        All commands will be translated by the given amount,
        including absolute location commands."""
        self.add_stitch_relative(MATRIX_TRANSLATE, x, y)

    @staticmethod
    def supported_formats():
        """Generates dictionary entries for supported formats. Each entry will
        always have description, extension, mimetype, and category. Reader
        will provide the reader, if one exists, writer will provide the writer,
        if one exists.

        Metadata gives a list of metadata read and/or written by that type.

        Options provides accepted options by the format and their accepted values.
        """
        # yield ({
        #     "description": "Art Embroidery Format",
        #     "extension": "art",
        #     "extensions": ("art",),
        #     "mimetype": "application/x-art",
        #     "category": "embroidery",
        #     "reader": ArtReader,
        #     "metadata": ("name")
        # })
        yield ({
            "description": "Brother Embroidery Format",
            "extension": "pec",
            "extensions": ("pec",),
            "mimetype": "application/x-pec",
            "category": "embroidery",
            "reader": PecReader,
            "writer": PecWriter,
            "metadata": ("name")
        })
        yield ({
            "description": "Brother Embroidery Format",
            "extension": "pes",
            "extensions": ("pes",),
            "mimetype": "application/x-pes",
            "category": "embroidery",
            "reader": PesReader,
            "writer": PesWriter,
            "versions": ("1", "6", "1t", "6t"),
            "metadata": ("name", "author", "category", "keywords", "comments")
        })
        yield ({
            "description": "Melco Embroidery Format",
            "extension": "exp",
            "extensions": ("exp",),
            "mimetype": "application/x-exp",
            "category": "embroidery",
            "reader": ExpReader,
            "writer": ExpWriter,
        })
        yield ({
            "description": "Tajima Embroidery Format",
            "extension": "dst",
            "extensions": ("dst",),
            "mimetype": "application/x-dst",
            "category": "embroidery",
            "reader": DstReader,
            "writer": DstWriter,
            "read_options": {
                "trim_distance": (None, 3.0, 50.0),
                "trim_at": (2, 3, 4, 5, 6, 7, 8),
                "clipping": (True, False)
            },
            "write_options": {
                "trim_at": (2, 3, 4, 5, 6, 7, 8)
            },
            "versions": ("default", "extended"),
            "metadata": ("name", "author", "copyright")
        })
        yield ({
            "description": "Janome Embroidery Format",
            "extension": "jef",
            "extensions": ("jef",),
            "mimetype": "application/x-jef",
            "category": "embroidery",
            "reader": JefReader,
            "writer": JefWriter,
            "read_options": {
                "trim_distance": (None, 3.0, 50.0),
                "trims": (True, False),
                "trim_at": (2, 3, 4, 5, 6, 7, 8),
                "clipping": (True, False)
            },
            "write_options": {
                "trims": (True, False),
                "trim_at": (2, 3, 4, 5, 6, 7, 8),
            },
        })
        yield ({
            "description": "Pfaff Embroidery Format",
            "extension": "vp3",
            "extensions": ("vp3",),
            "mimetype": "application/x-vp3",
            "category": "embroidery",
            "reader": Vp3Reader,
            "writer": Vp3Writer,
        })
        yield ({
            "description": "Scalable Vector Graphics",
            "extension": "svg",
            "extensions": ("svg", "svgz"),
            "mimetype": "image/svg+xml",
            "category": "vector",
            "writer": SvgWriter,
        })
        yield ({
            "description": "Comma-separated values",
            "extension": "csv",
            "extensions": ("csv",),
            "mimetype": "text/csv",
            "category": "debug",
            "reader": CsvReader,
            "writer": CsvWriter,
            "versions": ("default", "delta", "full")
        })
        yield ({
            "description": "Singer Embroidery Format",
            "extension": "xxx",
            "extensions": ("xxx",),
            "mimetype": "application/x-xxx",
            "category": "embroidery",
            "reader": XxxReader,
            "writer": XxxWriter
        })
        yield ({
            "description": "Janome Embroidery Format",
            "extension": "sew",
            "extensions": ("sew",),
            "mimetype": "application/x-sew",
            "category": "embroidery",
            "reader": SewReader
        })
        yield ({
            "description": "Barudan Embroidery Format",
            "extension": "u01",
            "extensions": ("u00", "u01", "u02"),
            "mimetype": "application/x-u01",
            "category": "embroidery",
            "reader": U01Reader,
            "writer": U01Writer
        })
        yield ({
            "description": "Husqvarna Viking Embroidery Format",
            "extension": "shv",
            "extensions": ("shv",),
            "mimetype": "application/x-shv",
            "category": "embroidery",
            "reader": ShvReader
        })
        yield ({
            "description": "Toyota Embroidery Format",
            "extension": "10o",
            "extensions": ("10o",),
            "mimetype": "application/x-10o",
            "category": "embroidery",
            "reader": A10oReader
        })
        yield ({
            "description": "Toyota Embroidery Format",
            "extension": "100",
            "extensions": ("100",),
            "mimetype": "application/x-100",
            "category": "embroidery",
            "reader": A100Reader
        })
        yield ({
            "description": "Bits & Volts Embroidery Format",
            "extension": "bro",
            "extensions": ("bro",),
            "mimetype": "application/x-Bro",
            "category": "embroidery",
            "reader": BroReader
        })
        yield ({
            "description": "Sunstar or Barudan Embroidery Format",
            "extension": "dat",
            "extensions": ("dat",),
            "mimetype": "application/x-dat",
            "category": "embroidery",
            "reader": DatReader
        })
        yield ({
            "description": "Tajima(Barudan) Embroidery Format",
            "extension": "dsb",
            "extensions": ("dsb",),
            "mimetype": "application/x-dsb",
            "category": "embroidery",
            "reader": DsbReader
        })
        yield ({
            "description": "ZSK USA Embroidery Format",
            "extension": "dsz",
            "extensions": ("dsz",),
            "mimetype": "application/x-dsz",
            "category": "embroidery",
            "reader": DszReader
        })
        yield ({
            "description": "Elna Embroidery Format",
            "extension": "emd",
            "extensions": ("emd",),
            "mimetype": "application/x-emd",
            "category": "embroidery",
            "reader": EmdReader
        })
        yield ({
            "description": "Eltac Embroidery Format",
            "extension": "exy",  # e??, e01
            "extensions": ("e00", "e01", "e02"),
            "mimetype": "application/x-exy",
            "category": "embroidery",
            "reader": ExyReader
        })
        yield ({
            "description": "Fortron Embroidery Format",
            "extension": "fxy",  # f??, f01
            "extensions": ("f00", "f01", "f02"),
            "mimetype": "application/x-fxy",
            "category": "embroidery",
            "reader": FxyReader
        })
        yield ({
            "description": "Gold Thread Embroidery Format",
            "extension": "gt",
            "extensions": ("gt",),
            "mimetype": "application/x-exy",
            "category": "embroidery",
            "reader": GtReader
        })
        yield ({
            "description": "Inbro Embroidery Format",
            "extension": "inb",
            "extensions": ("inb",),
            "mimetype": "application/x-inb",
            "category": "embroidery",
            "reader": InbReader
        })
        yield ({
            "description": "Tajima Embroidery Format",
            "extension": "tbf",
            "extensions": ("tbf",),
            "mimetype": "application/x-tbf",
            "category": "embroidery",
            "reader": TbfReader
        })
        yield ({
            "description": "Pfaff Embroidery Format",
            "extension": "ksm",
            "extensions": ("ksm",),
            "mimetype": "application/x-ksm",
            "category": "embroidery",
            "reader": KsmReader
        })
        yield ({
            "description": "Happy Embroidery Format",
            "extension": "tap",
            "extensions": ("tap",),
            "mimetype": "application/x-tap",
            "category": "embroidery",
            "reader": TapReader
        })
        yield ({
            "description": "Pfaff Embroidery Format",
            "extension": "spx",
            "extensions": ("spx"),
            "mimetype": "application/x-spx",
            "category": "embroidery",
            "reader": SpxReader
        })
        yield ({
            "description": "Data Stitch Embroidery Format",
            "extension": "stx",
            "extensions": ("stx",),
            "mimetype": "application/x-stx",
            "category": "embroidery",
            "reader": StxReader
        })
        yield ({
            "description": "Brother Embroidery Format",
            "extension": "phb",
            "extensions": ("phb",),
            "mimetype": "application/x-phb",
            "category": "embroidery",
            "reader": PhbReader
        })
        yield ({
            "description": "Brother Embroidery Format",
            "extension": "phc",
            "extensions": ("phc",),
            "mimetype": "application/x-phc",
            "category": "embroidery",
            "reader": PhcReader
        })
        yield ({
            "description": "Ameco Embroidery Format",
            "extension": "new",
            "extensions": ("new",),
            "mimetype": "application/x-new",
            "category": "embroidery",
            "reader": NewReader
        })
        yield ({
            "description": "Pfaff Embroidery Format",
            "extension": "max",
            "extensions": ("max",),
            "mimetype": "application/x-max",
            "category": "embroidery",
            "reader": MaxReader
        })
        yield ({
            "description": "Mitsubishi Embroidery Format",
            "extension": "mit",
            "extensions": ("mit",),
            "mimetype": "application/x-mit",
            "category": "embroidery",
            "reader": MitReader
        })
        yield ({
            "description": "Pfaff Embroidery Format",
            "extension": "pcd",
            "extensions": ("pcd",),
            "mimetype": "application/x-pcd",
            "category": "embroidery",
            "reader": PcdReader
        })
        yield ({
            "description": "Pfaff Embroidery Format",
            "extension": "pcq",
            "extensions": ("pcq",),
            "mimetype": "application/x-pcq",
            "category": "embroidery",
            "reader": PcqReader
        })
        yield ({
            "description": "Pfaff Embroidery Format",
            "extension": "pcm",
            "extensions": ("pcm",),
            "mimetype": "application/x-pcm",
            "category": "embroidery",
            "reader": PcmReader
        })
        yield ({
            "description": "Pfaff Embroidery Format",
            "extension": "pcs",
            "extensions": ("pcs",),
            "mimetype": "application/x-pcs",
            "category": "embroidery",
            "reader": PcsReader
        })
        yield ({
            "description": "Janome Embroidery Format",
            "extension": "jpx",
            "extensions": ("jpx",),
            "mimetype": "application/x-jpx",
            "category": "embroidery",
            "reader": JpxReader
        })
        yield ({
            "description": "Gunold Embroidery Format",
            "extension": "stc",
            "extensions": ("stc",),
            "mimetype": "application/x-stc",
            "category": "embroidery",
            "reader": StcReader
        })
        # yield ({
        #     "description": "Zeng Hsing Embroidery Format",
        #     "extension": "zhs",
        #     "mimetype": "application/x-zhs",
        #     "category": "embroidery",
        #     "reader": ZhsReader
        # })
        yield ({
            "description": "ZSK TC Embroidery Format",
            "extension": "zxy",
            "extensions": ("z00", "z01", "z02"),
            "mimetype": "application/x-zxy",
            "category": "embroidery",
            "reader": ZxyReader
        })
        yield ({
            "description": "Brother Stitch Format",
            "extension": "pmv",
            "extensions": ("pmv",),
            "mimetype": "application/x-pmv",
            "category": "stitch",
            "reader": PmvReader,
            "writer": PmvWriter
        })
        yield ({
            "description": "PNG Format, Portable Network Graphics",
            "extension": "png",
            "extensions": ("png",),
            "mimetype": "image/png",
            "category": "image",
            "writer": PngWriter,
            "write_options": {
                "background": (0x000000, 0xFFFFFF),
                "linewidth": (1, 2, 3, 4, 5, 6, 7, 8, 9, 10)
            },
        })
        yield ({
            "description": "txt Format, Text File",
            "extension": "txt",
            "extensions": ("txt",),
            "mimetype": "text/plain",
            "category": "debug",
            "writer": TxtWriter,
            "versions": ("default", "embroidermodder")
        })
        yield ({
            "description": "gcode Format, Text File",
            "extension": "gcode",
            "extensions": ("gcode", "g-code", "ngc", "nc", ".g"),
            "mimetype": "text/plain",
            "category": "embroidery",
            "reader": GcodeReader,
            "writer": GcodeWriter,
            "write_options": {
                "stitch_z_travel": (5.0, 10.0),
            },
        })
        yield ({
            "description": "Husqvarna Embroidery Format",
            "extension": "hus",
            "extensions": ("hus",),
            "mimetype": "application/x-hus",
            "category": "embroidery",
            "reader": HusReader
        })
        yield ({
            "description": "Edr Color Format",
            "extension": "edr",
            "extensions": ("edr",),
            "mimetype": "application/x-edr",
            "category": "color",
            "reader": EdrReader,
            "writer": EdrWriter
        })
        yield ({
            "description": "Col Color Format",
            "extension": "col",
            "extensions": ("col",),
            "mimetype": "application/x-col",
            "category": "color",
            "reader": ColReader,
            "writer": ColWriter
        })
        yield ({
            "description": "Inf Color Format",
            "extension": "inf",
            "extensions": ("inf",),
            "mimetype": "application/x-inf",
            "category": "color",
            "reader": InfReader,
            "writer": InfWriter
        })
        yield ({
            "description": "Json Export",
            "extension": "json",
            "extensions": ("json",),
            "mimetype": "application/json",
            "category": "debug",
            "reader": JsonReader,
            "writer": JsonWriter
        })

    @staticmethod
    def convert(filename_from, filename_to, settings=None):
        pattern = EmbPattern.static_read(filename_from, settings)
        if pattern is None:
            return
        EmbPattern.static_write(pattern, filename_to, settings)

    @staticmethod
    def get_extension_by_filename(filename):
        """extracts the extension from a filename"""
        return os.path.splitext(filename)[1][1:]

    @staticmethod
    def read_embroidery(reader, f, settings=None, pattern=None):
        """Reads fileobject or filename with reader."""
        if reader is None:
            return None
        if pattern is None:
            pattern = EmbPattern()

        if EmbPattern.is_str(f):
            text_mode = False
            try:
                text_mode = reader.READ_FILE_IN_TEXT_MODE
            except AttributeError:
                pass
            if text_mode:
                try:
                    with open(f, "r") as stream:
                        reader.read(stream, pattern, settings)
                        stream.close()
                except IOError:
                    pass
            else:
                try:
                    with open(f, "rb") as stream:
                        reader.read(stream, pattern, settings)
                        stream.close()
                except IOError:
                    pass
        else:
            reader.read(f, pattern, settings)
        return pattern

    @staticmethod
    def read_dst(f, settings=None, pattern=None):
        """Reads fileobject as DST file"""
        return EmbPattern.read_embroidery(DstReader, f, settings, pattern)

    @staticmethod
    def read_pec(f, settings=None, pattern=None):
        """Reads fileobject as PEC file"""
        return EmbPattern.read_embroidery(PecReader, f, settings, pattern)

    @staticmethod
    def read_pes(f, settings=None, pattern=None):
        """Reads fileobject as PES file"""
        return EmbPattern.read_embroidery(PesReader, f, settings, pattern)

    @staticmethod
    def read_exp(f, settings=None, pattern=None):
        """Reads fileobject as EXP file"""
        return EmbPattern.read_embroidery(ExpReader, f, settings, pattern)

    @staticmethod
    def read_vp3(f, settings=None, pattern=None):
        """Reads fileobject as VP3 file"""
        return EmbPattern.read_embroidery(Vp3Reader, f, settings, pattern)

    @staticmethod
    def read_jef(f, settings=None, pattern=None):
        """Reads fileobject as JEF file"""
        return EmbPattern.read_embroidery(JefReader, f, settings, pattern)

    @staticmethod
    def read_u01(f, settings=None, pattern=None):
        """Reads fileobject as U01 file"""
        return EmbPattern.read_embroidery(U01Reader, f, settings, pattern)

    @staticmethod
    def read_csv(f, settings=None, pattern=None):
        """Reads fileobject as CSV file"""
        return EmbPattern.read_embroidery(CsvReader, f, settings, pattern)

    @staticmethod
    def read_gcode(f, settings=None, pattern=None):
        """Reads fileobject as GCode file"""
        return EmbPattern.read_embroidery(GcodeReader, f, settings, pattern)

    @staticmethod
    def read_xxx(f, settings=None, pattern=None):
        """Reads fileobject as XXX file"""
        return EmbPattern.read_embroidery(XxxReader, f, settings, pattern)

    @staticmethod
    def static_read(filename, settings=None, pattern=None):
        """Reads file, assuming type by extension"""
        extension = EmbPattern.get_extension_by_filename(filename)
        extension = extension.lower()
        for file_type in EmbPattern.supported_formats():
            if file_type['extension'] != extension:
                continue
            reader = file_type.get("reader", None)
            return EmbPattern.read_embroidery(reader, filename, settings, pattern)
        return None

    @staticmethod
    def write_embroidery(writer, pattern, stream, settings=None):
        if pattern is None:
            return
        if settings is None:
            settings = {}
        else:
            settings = settings.copy()
        try:
            encode = writer.ENCODE
        except AttributeError:
            encode = True

        if settings.get("encode", encode):
            if not ("max_jump" in settings):
                try:
                    settings["max_jump"] = writer.MAX_JUMP_DISTANCE
                except AttributeError:
                    pass
            if not ("max_stitch" in settings):
                try:
                    settings["max_stitch"] = writer.MAX_STITCH_DISTANCE
                except AttributeError:
                    pass
            if not ("full_jump" in settings):
                try:
                    settings["full_jump"] = writer.FULL_JUMP
                except AttributeError:
                    pass
            if not ("round" in settings):
                try:
                    settings["round"] = writer.ROUND
                except AttributeError:
                    pass
            if not ("writes_speeds" in settings):
                try:
                    settings["writes_speeds"] = writer.WRITES_SPEEDS
                except AttributeError:
                    pass
            if not ("sequin_contingency" in settings):
                try:
                    settings["sequin_contingency"] = writer.SEQUIN_CONTINGENCY
                except AttributeError:
                    pass
            if not ("thread_change_command" in settings):
                try:
                    settings["thread_change_command"] = writer.THREAD_CHANGE_COMMAND
                except AttributeError:
                    pass
            if not ("translate" in settings):
                try:
                    settings["translate"] = writer.TRANSLATE
                except AttributeError:
                    pass
            if not ("scale" in settings):
                try:
                    settings["scale"] = writer.SCALE
                except AttributeError:
                    pass
            if not ("rotate" in settings):
                try:
                    settings["rotate"] = writer.ROTATE
                except AttributeError:
                    pass
            pattern = pattern.get_normalized_pattern(settings)

        if EmbPattern.is_str(stream):
            text_mode = False
            try:
                text_mode = writer.WRITE_FILE_IN_TEXT_MODE
            except AttributeError:
                pass
            if text_mode:
                try:
                    with open(stream, "w") as stream:
                        writer.write(pattern, stream, settings)
                except IOError:
                    pass
            else:
                try:
                    with open(stream, "wb") as stream:
                        writer.write(pattern, stream, settings)
                except IOError:
                    pass
        else:
            writer.write(pattern, stream, settings)

    @staticmethod
    def write_dst(pattern, stream, settings=None):
        """Writes fileobject as DST file"""
        EmbPattern.write_embroidery(DstWriter, pattern, stream, settings)

    @staticmethod
    def write_pec(pattern, stream, settings=None):
        """Writes fileobject as PEC file"""
        EmbPattern.write_embroidery(PecWriter, pattern, stream, settings)

    @staticmethod
    def write_pes(pattern, stream, settings=None):
        """Writes fileobject as PES file"""
        EmbPattern.write_embroidery(PesWriter, pattern, stream, settings)

    @staticmethod
    def write_exp(pattern, stream, settings=None):
        """Writes fileobject as EXP file"""
        EmbPattern.write_embroidery(ExpWriter, pattern, stream, settings)

    @staticmethod
    def write_vp3(pattern, stream, settings=None):
        """Writes fileobject as Vp3 file"""
        EmbPattern.write_embroidery(Vp3Writer, pattern, stream, settings)

    @staticmethod
    def write_jef(pattern, stream, settings=None):
        """Writes fileobject as JEF file"""
        EmbPattern.write_embroidery(JefWriter, pattern, stream, settings)

    @staticmethod
    def write_u01(pattern, stream, settings=None):
        """Writes fileobject as U01 file"""
        EmbPattern.write_embroidery(U01Writer, pattern, stream, settings)

    @staticmethod
    def write_csv(pattern, stream, settings=None):
        """Writes fileobject as CSV file"""
        EmbPattern.write_embroidery(CsvWriter, pattern, stream, settings)

    @staticmethod
    def write_txt(pattern, stream, settings=None):
        """Writes fileobject as CSV file"""
        EmbPattern.write_embroidery(TxtWriter, pattern, stream, settings)

    @staticmethod
    def write_gcode(pattern, stream, settings=None):
        """Writes fileobject as Gcode file"""
        EmbPattern.write_embroidery(GcodeWriter, pattern, stream, settings)

    @staticmethod
    def write_xxx(pattern, stream, settings=None):
        """Writes fileobject as XXX file"""
        EmbPattern.write_embroidery(XxxWriter, pattern, stream, settings)

    @staticmethod
    def write_svg(pattern, stream, settings=None):
        """Writes fileobject as DST file"""
        EmbPattern.write_embroidery(SvgWriter, pattern, stream, settings)

    @staticmethod
    def write_png(pattern, stream, settings=None):
        """Writes fileobject as PNG file"""
        EmbPattern.write_embroidery(PngWriter, pattern, stream, settings)

    @staticmethod
    def static_write(pattern, filename, settings=None):
        """Writes file, assuming type by extension"""
        extension = EmbPattern.get_extension_by_filename(filename)
        extension = extension.lower()

        for file_type in EmbPattern.supported_formats():
            if file_type['extension'] != extension:
                continue
            writer = file_type.get("writer", None)
            if writer is None:
                continue
            EmbPattern.write_embroidery(writer, pattern, filename, settings)

    @staticmethod
    def is_str(obj):
        try:
            return isinstance(obj, basestring)
        except NameError:
            return isinstance(obj, str)



from .EmbThreadPec import get_thread_set
from .PecReader import read_pec_stitches, read_pec_graphics
from .ReadHelper import read_int_8, read_int_32le, read_int_16le


def read(f, out, settings=None):
    f.seek(0x4A, 0)
    pec_graphic_icon_height = read_int_8(f)
    f.seek(1, 1)
    pec_graphic_byte_stride = read_int_8(f)
    color_count = read_int_16le(f)
    threadset = get_thread_set()
    for i in range(0, color_count):
        color_index = read_int_8(f)
        if color_index is None:
            return  # File terminated before expected end.
        out.add_thread(threadset[color_index % len(threadset)])
    byte_size = pec_graphic_byte_stride * pec_graphic_icon_height
    read_pec_graphics(f,
                      out,
                      byte_size,
                      pec_graphic_byte_stride,
                      color_count,
                      out.threadlist
                      )
    f.seek(0x2B, 0)
    pec_add = read_int_8(f)  # Size of pre-graphics, post copyright header.
    f.seek(4, 1)  # 0x30, graphics end size.
    pec_offset = read_int_16le(f)
    f.seek(pec_offset + pec_add, 0)
    bytes_in_section = read_int_16le(f)  # Primary bounds.
    f.seek(bytes_in_section, 1)
    bytes_in_section2 = read_int_32le(f)  # Sectional bounds.
    f.seek(bytes_in_section2 + 10, 1)
    color_count2 = read_int_8(f)
    f.seek(color_count2 + 0x1D, 1)  # 1D toto back
    read_pec_stitches(f, out)
    out.interpolate_duplicate_color_as_stop()




from .EmbFunctions import *

READ_FILE_IN_TEXT_MODE = True


def read(f, out, settings=None):
    import csv
    csv_reader = csv.reader(f, delimiter=',')
    command_dict = get_command_dictionary()
    for row in csv_reader:
        if len(row) == 0:
            continue
        if "*" in row[0]:
            split = row[2].split(' ')
            command = command_dict[split[0]]
            for sp in split[1:]:
                if sp[0] == "n":
                    needle = int(sp[1:])
                    command |= (needle + 1) << 16
                if sp[0] == "o":
                    order = int(sp[1:])
                    command |= (order + 1) << 24
                if sp[0] == "t":
                    thread = int(sp[1:])
                    command |= (thread + 1) << 8
            if len(row) == 3:
                out.add_command(command)
            else:
                out.add_stitch_absolute(command, float(row[3]), float(row[4]))
        elif "#" in row[0]:
            continue
        elif "@" in row[0]:
            if len(row) != 3:
                continue
            out.metadata(row[1], row[2])
        elif "$" in row[0]:
            thread_add = {}
            if len(row) == 7 and len(row[2]) <= 3 and len(row[3]) <= 3 and len(row[4]) <= 3:
                # This is an embroidermodder csv file, I changed the colors and added more details.
                # [THREAD_NUMBER], [RED], [GREEN], [BLUE], [DESCRIPTION], [CATALOG_NUMBER]\"\n");
                thread_add["rgb"] = (int(row[2]), int(row[3]), int(row[4]))
                thread_add["description"] = row[5]
                thread_add["catalog"] = row[6]
            else:
                try:
                    thread_add["rgb"] = row[2]
                except IndexError:
                    pass
                try:
                    thread_add["name"] = row[3]
                except IndexError:
                    pass
                try:
                    thread_add["brand"] = row[4]
                except IndexError:
                    pass
                try:
                    thread_add["catalog"] = row[5]
                except IndexError:
                    pass
                try:
                    thread_add["details"] = row[6]
                except IndexError:
                    pass
                try:
                    thread_add["weight"] = row[7]
                except IndexError:
                    pass
            out.add_thread(thread_add)



import math

from .EmbFunctions import *
from .PecGraphics import get_graphic_as_string
from .WriteHelper import write_string_utf8

ENCODE = False
WRITES_SPEEDS = True
SEQUIN_CONTINGENCY = CONTINGENCY_SEQUIN_UTILIZE


def csv(f, values):
    string = ""
    for v in values:
        if len(string) > 0:
            string += ','
        string += ('\"%s\"' % v)
    write_string_utf8(f, string + "\n")


def distance(dx, dy):
    dx *= dx
    dy *= dy
    return math.sqrt(dx + dy)


def angle(dx, dy):
    tau = math.pi * 2
    angle = math.atan2(dy, dx)
    angle += tau / float(2)
    angle /= tau
    return angle


def write_data(pattern, f):
    names = get_common_name_dictionary()
    extends = pattern.bounds()
    width = extends[2] - extends[0]
    height = extends[3] - extends[1]

    csv(f, ('#', '[VAR_NAME]', '[VAR_VALUE]'))
    count_stitches = pattern.count_stitches()
    csv(f, ('>', 'STITCH_COUNT:', str(count_stitches)))
    count_threads = pattern.count_color_changes()
    csv(f, ('>', 'THREAD_COUNT:', str(count_threads)))
    count_set_needles = pattern.count_needle_sets()
    csv(f, ('>', 'NEEDLE_COUNT:', str(count_set_needles)))
    csv(f, ('>', 'EXTENTS_LEFT:', str(extends[0])))
    csv(f, ('>', 'EXTENTS_TOP:', str(extends[1])))
    csv(f, ('>', 'EXTENTS_RIGHT:', str(extends[2])))
    csv(f, ('>', 'EXTENTS_BOTTOM:', str(extends[3])))
    csv(f, ('>', 'EXTENTS_WIDTH:', str(width)))
    csv(f, ('>', 'EXTENTS_HEIGHT:', str(height)))

    stitch_counts = {}
    for s in pattern.stitches:
        command = s[2] & COMMAND_MASK
        if command in stitch_counts:
            stitch_counts[command] += 1
        else:
            stitch_counts[command] = 1

    if len(stitch_counts) != 0:
        for the_key, the_value in stitch_counts.items():
            try:
                the_key &= COMMAND_MASK
                name = "COMMAND_" + names[the_key]
            except (IndexError, KeyError):
                name = "COMMAND_UNKNOWN_" + str(the_key)
            csv(f, (
                '>',
                name,
                str(the_value)
            ))

    write_string_utf8(f, "\n")


def write_metadata(pattern, f):
    if len(pattern.extras) > 0:
        csv(f, (
            '#',
            '[METADATA_NAME]',
            '[METADATA]'
        ))
        for the_key, the_value in pattern.extras.items():
            if isinstance(the_value, tuple):
                the_value = "\n" + get_graphic_as_string(the_value)
            csv(f, (
                '@',
                str(the_key),
                str(the_value)
            ))
        write_string_utf8(f, "\n")


def write_threads(pattern, f):
    if len(pattern.threadlist) > 0:
        csv(f, (
            '#',
            '[THREAD_NUMBER]',
            '[HEX_COLOR]',
            '[DESCRIPTION]',
            '[BRAND]',
            '[CATALOG_NUMBER]',
            '[DETAILS]',
            '[WEIGHT]'
        ))
        for i, thread in enumerate(pattern.threadlist):
            csv(f, (
                '$',
                str(i),
                thread.hex_color(),
                thread.description,
                thread.brand,
                thread.catalog_number,
                thread.details,
                thread.weight,
            ))
        write_string_utf8(f, "\n")


def decoded_name(names, data):
    command = decode_embroidery_command(data)
    try:
        name = names[command[0]]
        if command[1] is not None:
            name = name + " t" + str(command[1])
        if command[2] is not None:
            name = name + " n" + str(command[2])
        if command[3] is not None:
            name = name + " o" + str(command[3])
    except (IndexError, KeyError):
        name = "UNKNOWN " + str(data)
    return name


def write_stitches_displacement(pattern, f):
    names = get_common_name_dictionary()
    csv(f, (
        '#',
        '[STITCH_INDEX]',
        '[STITCH_TYPE]',
        '[X]',
        '[Y]',
        '[DX]',
        '[DY]',
        '[R]',
        '[ANGLE]'
    ))

    current_x = 0
    current_y = 0
    for i, stitch in enumerate(pattern.stitches):
        name = decoded_name(names, stitch[2])
        dx = stitch[0] - current_x
        dy = stitch[1] - current_y
        csv(f, (
            '*',
            str(i),
            name,
            str(stitch[0]),
            str(stitch[1]),
            str(dx),
            str(dy),
            str(distance(dx, dy)),
            str(angle(dx, dy))
        ))
        current_x = stitch[0]
        current_y = stitch[1]


def write_stitches_deltas(pattern, f):
    names = get_common_name_dictionary()
    csv(f, (
        '#',
        '[STITCH_INDEX]',
        '[STITCH_TYPE]',
        '[X]',
        '[Y]',
        '[DX]',
        '[DY]'
    ))
    current_x = 0
    current_y = 0
    for i, stitch in enumerate(pattern.stitches):
        name = decoded_name(names, stitch[2])
        dx = stitch[0] - current_x
        dy = stitch[1] - current_y
        csv(f, (
            '*',
            str(i),
            name,
            str(stitch[0]),
            str(stitch[1]),
            str(dx),
            str(dy)
        ))
        current_x = stitch[0]
        current_y = stitch[1]


def write_stitches(pattern, f):
    names = get_common_name_dictionary()
    csv(f, (
        '#',
        '[STITCH_INDEX]',
        '[STITCH_TYPE]',
        '[X]',
        '[Y]'
    ))
    for i, stitch in enumerate(pattern.stitches):
        name = decoded_name(names, stitch[2])
        csv(f, (
            '*',
            str(i),
            name,
            str(stitch[0]),
            str(stitch[1]),
        ))


def write(pattern, f, settings=None):
    version = "default"
    if settings is not None:
        if "deltas" in settings:
            version = "delta"
        elif "displacement" in settings:
            version = "full"
        version = settings.get("version", version)
    write_data(pattern, f)
    write_metadata(pattern, f)
    write_threads(pattern, f)

    if len(pattern.stitches) > 0:
        if version == "full":
            write_stitches_displacement(pattern, f)
        elif version == "delta":
            write_stitches_deltas(pattern, f)
        else:
            write_stitches(pattern, f)



from .EmbFunctions import *
from .WriteHelper import write_string_utf8


def write_mimic(pattern, f):
    emb_mod_convert = {
        STITCH: 0,
        JUMP: 1,
        TRIM: 2,
        STOP: 4,
        COLOR_CHANGE: 4,
        NEEDLE_SET: 4,
        SEQUIN_MODE: 8,
        SEQUIN_EJECT: 8,
        END: 16,
        # define NORMAL              0 /* stitch to (xx, yy) */
        # define JUMP                1 /* move to(xx, yy) */
        # define TRIM                2 /* trim + move to(xx, yy) */
        # define STOP                4 /* pause machine for thread change */
        # define SEQUIN              8 /* sequin */
        # define END                 16 /* end of program */
    }
    color = 0
    for i, stitch in enumerate(pattern.stitches):
        xx = stitch[0]
        yy = stitch[1]
        flags = stitch[2]
        if (flags & COMMAND_MASK) == COLOR_CHANGE:
            color += 1
        flags = emb_mod_convert[(flags & COMMAND_MASK)]
        txt_line = "%.1f,%.1f color:%i flags:%s\n" % (xx / 10.0, yy / 10.0, color, flags)
        write_string_utf8(f, txt_line)


def write_normal(pattern, f):
    names = get_common_name_dictionary()
    color_index = 0
    color = pattern.get_thread_or_filler(color_index).color
    for i, stitch in enumerate(pattern.stitches):
        xx = stitch[0]
        yy = stitch[1]
        flags = stitch[2] & COMMAND_MASK
        if (flags & COMMAND_MASK) == COLOR_CHANGE:
            color = pattern.get_thread_or_filler(color_index).color
        named_command = names[(flags & COMMAND_MASK)]
        txt_line = "%.1f,%.1f color:%i command:%s flags:%i\n" % (xx, yy, color, named_command, flags)
        write_string_utf8(f, txt_line)


def write(pattern, f, settings=None):
    mimic = False
    if settings is not None:
        if "mimic" in settings:
            mimic = True
        version = settings.get("version", "default")
        if version != "default":
            mimic = True
    if mimic:
        write_mimic(pattern, f)
    else:
        write_normal(pattern, f)



from .EmbThread import EmbThread
from .PecReader import read_pec
from .ReadHelper import read_string_8, read_int_8, read_int_32le, read_int_24be, read_int_16le


def read(f, out, settings=None):
    loaded_thread_values = []
    pes_string = read_string_8(f, 8)

    if pes_string == "#PEC0001":
        read_pec(f, out, loaded_thread_values)
        out.interpolate_duplicate_color_as_stop()
        return

    pec_block_position = read_int_32le(f)

    # Ignoring several known PES versions, just abort and read PEC block
    # All versions allow, abort and read PEC block.
    # Metadata started appearing in V4
    # Threads appeared in V5.
    # We quickly abort if there are any complex items in the header.
    if pes_string == "#PES0100":
        out.metadata("version", 10)
        read_pes_header_version_10(f, out, loaded_thread_values)
    elif pes_string == "#PES0090":
        out.metadata("version", 9)
        read_pes_header_version_9(f, out, loaded_thread_values)
    elif pes_string == "#PES0080":
        out.metadata("version", 8)
        read_pes_header_version_8(f, out, loaded_thread_values)
    elif pes_string == "#PES0070":
        out.metadata("version", 7)
        read_pes_header_version_7(f, out, loaded_thread_values)
    elif pes_string == "#PES0060":
        out.metadata("version", 6)
        read_pes_header_version_6(f, out, loaded_thread_values)
    elif pes_string == "#PES0050":
        out.metadata("version", 5)
        read_pes_header_version_5(f, out, loaded_thread_values)
    elif pes_string == "#PES0055":
        out.metadata("version", 5.5)
        read_pes_header_version_5(f, out, loaded_thread_values)
    elif pes_string == "#PES0056":
        out.metadata("version", 5.6)
        read_pes_header_version_5(f, out, loaded_thread_values)
    elif pes_string == "#PES0040":
        out.metadata("version", 4)
        read_pes_header_version_4(f, out)
    elif pes_string == "#PES0030":
        out.metadata("version", 3)
    elif pes_string == "#PES0022":
        out.metadata("version", 2.2)
    elif pes_string == "#PES0020":
        out.metadata("version", 2)
    elif pes_string == "#PES0001":
        out.metadata("version", 1)
        read_pes_header_version_1(f, out)
    else:
        pass  # Header is unrecognised.
    f.seek(pec_block_position, 0)
    read_pec(f, out, loaded_thread_values)
    out.interpolate_duplicate_color_as_stop()


def read_pes_string(f):
    length = read_int_8(f)
    if length == 0:
        return None
    return read_string_8(f, length)


def read_pes_metadata(f, out):
    v = read_pes_string(f)
    if v is not None and len(v) > 0:
        out.metadata("name", v)
    v = read_pes_string(f)
    if v is not None and len(v) > 0:
        out.metadata("category", v)
    v = read_pes_string(f)
    if v is not None and len(v) > 0:
        out.metadata("author", v)
    v = read_pes_string(f)
    if v is not None and len(v) > 0:
        out.metadata("keywords", v)
    v = read_pes_string(f)
    if v is not None and len(v) > 0:
        out.metadata("comments", v)


def read_pes_thread(f, threadlist):
    thread = EmbThread()
    thread.catalog_number = read_pes_string(f)
    thread.color = 0xFF000000 | read_int_24be(f)
    f.seek(5, 1)
    thread.description = read_pes_string(f)
    thread.brand = read_pes_string(f)
    thread.chart = read_pes_string(f)
    threadlist.append(thread)


def read_pes_header_version_1(f, out):
    # Nothing I care about.
    pass


def read_pes_header_version_4(f, out):
    f.seek(4, 1)
    read_pes_metadata(f, out)


def read_pes_header_version_5(f, out, threadlist):
    f.seek(4, 1)
    read_pes_metadata(f, out)
    f.seek(24, 1)  # this is 36 in version 6 and 24 in version 5
    v = read_pes_string(f)
    if v is not None and len(v) > 0:
        out.metadata("image", v)
    f.seek(24, 1)
    count_programmable_fills = read_int_16le(f)
    if count_programmable_fills != 0:
        return
    count_motifs = read_int_16le(f)
    if count_motifs != 0:
        return
    count_feather_patterns = read_int_16le(f)
    if count_feather_patterns != 0:
        return
    count_threads = read_int_16le(f)
    for i in range(0, count_threads):
        read_pes_thread(f, threadlist)


def read_pes_header_version_6(f, out, threadlist):
    f.seek(4, 1)
    read_pes_metadata(f, out)
    f.seek(36, 1)  # this is 36 in version 6 and 24 in version 5
    v = read_pes_string(f)
    if v is not None and len(v) > 0:
        out.metadata("image_file", v)
    f.seek(24, 1)
    count_programmable_fills = read_int_16le(f)
    if count_programmable_fills != 0:
        return
    count_motifs = read_int_16le(f)
    if count_motifs != 0:
        return
    count_feather_patterns = read_int_16le(f)
    if count_feather_patterns != 0:
        return
    count_threads = read_int_16le(f)
    for i in range(0, count_threads):
        read_pes_thread(f, threadlist)


def read_pes_header_version_7(f, out, threadlist):
    f.seek(4, 1)
    read_pes_metadata(f, out)
    f.seek(36, 1)
    v = read_pes_string(f)
    if v is not None and len(v) > 0:
        out.metadata("image_file", v)
    f.seek(24, 1)
    count_programmable_fills = read_int_16le(f)
    if count_programmable_fills != 0:
        return
    count_motifs = read_int_16le(f)
    if count_motifs != 0:
        return
    count_feather_patterns = read_int_16le(f)
    if count_feather_patterns != 0:
        return
    count_threads = read_int_16le(f)
    for i in range(0, count_threads):
        read_pes_thread(f, threadlist)


def read_pes_header_version_8(f, out, threadlist):
    f.seek(4, 1)
    read_pes_metadata(f, out)
    f.seek(38, 1)
    v = read_pes_string(f)
    if v is not None and len(v) > 0:
        out.metadata("image_file", v)
    f.seek(26, 1)
    count_programmable_fills = read_int_16le(f)
    if count_programmable_fills != 0:
        return
    count_motifs = read_int_16le(f)
    if count_motifs != 0:
        return
    count_feather_patterns = read_int_16le(f)
    if count_feather_patterns != 0:
        return
    count_threads = read_int_16le(f)
    for i in range(0, count_threads):
        read_pes_thread(f, threadlist)


def read_pes_header_version_9(f, out, threadlist):
    f.seek(4, 1)
    read_pes_metadata(f, out)
    f.seek(14, 1)
    v = read_pes_string(f)
    if v is not None and len(v) > 0:
        out.metadata("hoop_name", v)
    f.seek(30, 1)  # this is 36 in version 6 and 24 in version 5
    v = read_pes_string(f)
    if v is not None and len(v) > 0:
        out.metadata("image_file", v)
    f.seek(34, 1)
    count_programmable_fills = read_int_16le(f)
    if count_programmable_fills != 0:
        return
    count_motifs = read_int_16le(f)
    if count_motifs != 0:
        return
    count_feather_patterns = read_int_16le(f)
    if count_feather_patterns != 0:
        return
    count_threads = read_int_16le(f)
    for i in range(0, count_threads):
        read_pes_thread(f, threadlist)


def read_pes_header_version_10(f, out, threadlist):
    f.seek(4, 1)
    read_pes_metadata(f, out)
    f.seek(14, 1)
    v = read_pes_string(f)
    if v is not None and len(v) > 0:
        out.metadata("hoop_name", v)
    f.seek(38, 1)
    v = read_pes_string(f)
    if v is not None and len(v) > 0:
        out.metadata("image_file", v)
    f.seek(34, 1)
    count_programmable_fills = read_int_16le(f)
    if count_programmable_fills != 0:
        return
    count_motifs = read_int_16le(f)
    if count_motifs != 0:
        return
    count_feather_patterns = read_int_16le(f)
    if count_feather_patterns != 0:
        return
    count_threads = read_int_16le(f)
    for i in range(0, count_threads):
        read_pes_thread(f, threadlist)



from .WriteHelper import write_string_utf8

ENCODE = False


def write(pattern, f, settings=None):
    write_string_utf8(f, "%d\r\n" % len(pattern.threadlist))
    index = 0
    for thread in pattern.threadlist:
        write_string_utf8(f, "%d,%d,%d,%d\r\n" % (
            index,
            thread.get_red(),
            thread.get_green(),
            thread.get_blue()))
        index += 1



from .EmbThread import EmbThread


def get_thread_set():
    return [
        EmbThreadShv(0, 0, 0, "Black", "0"),
        EmbThreadShv(0, 0, 255, "Blue", "1"),
        EmbThreadShv(51, 204, 102, "Green", "2"),
        EmbThreadShv(255, 0, 0, "Red", "3"),
        EmbThreadShv(255, 0, 255, "Purple", "4"),
        EmbThreadShv(255, 255, 0, "Yellow", "5"),
        EmbThreadShv(127, 127, 127, "Gray", "6"),
        EmbThreadShv(51, 154, 255, "Light Blue", "7"),
        EmbThreadShv(0, 255, 0, "Light Green", "8"),
        EmbThreadShv(255, 127, 0, "Orange", "9"),
        EmbThreadShv(255, 160, 180, "Pink", "10"),
        EmbThreadShv(153, 75, 0, "Brown", "11"),
        EmbThreadShv(255, 255, 255, "White", "12"),
        EmbThreadShv(0, 0, 0, "Black", "13"),
        EmbThreadShv(0, 0, 0, "Black", "14"),
        EmbThreadShv(0, 0, 0, "Black", "15"),
        EmbThreadShv(0, 0, 0, "Black", "16"),
        EmbThreadShv(0, 0, 0, "Black", "17"),
        EmbThreadShv(0, 0, 0, "Black", "18"),
        EmbThreadShv(255, 127, 127, "Light Red", "19"),
        EmbThreadShv(255, 127, 255, "Light Purple", "20"),
        EmbThreadShv(255, 255, 153, "Light Yellow", "21"),
        EmbThreadShv(192, 192, 192, "Light Gray", "22"),
        EmbThreadShv(0, 0, 0, "Black", "23"),
        EmbThreadShv(0, 0, 0, "Black", "24"),
        EmbThreadShv(255, 165, 65, "Light Orange", "25"),
        EmbThreadShv(255, 204, 204, "Light Pink", "26"),
        EmbThreadShv(175, 90, 10, "Light Brown", "27"),
        EmbThreadShv(0, 0, 0, "Black", "28"),
        EmbThreadShv(0, 0, 0, "Black", "29"),
        EmbThreadShv(0, 0, 0, "Black", "30"),
        EmbThreadShv(0, 0, 0, "Black", "31"),
        EmbThreadShv(0, 0, 0, "Black", "32"),
        EmbThreadShv(0, 0, 127, "Dark Blue", "33"),
        EmbThreadShv(0, 127, 0, "Dark Green", "34"),
        EmbThreadShv(127, 0, 0, "Dark Red", "35"),
        EmbThreadShv(127, 0, 127, "Dark Purple", "36"),
        EmbThreadShv(200, 200, 0, "Dark Yellow", "37"),
        EmbThreadShv(60, 60, 60, "Dark Gray", "38"),
        EmbThreadShv(0, 0, 0, "Black", "39"),
        EmbThreadShv(0, 0, 0, "Black", "40"),
        EmbThreadShv(232, 63, 0, "Dark Orange", "41"),
        EmbThreadShv(255, 102, 122, "Dark Pink", "42")
    ]


class EmbThreadShv(EmbThread):
    def __init__(self, red, green, blue, description, catalog_number):
        EmbThread.__init__(self)
        self.set_color(red, green, blue)
        self.description = description
        self.catalog_number = catalog_number
        self.brand = "Shv"
        self.chart = "Shv"



from .EmbThread import *

READ_FILE_IN_TEXT_MODE = True


def read(f, out, settings=None):
    count = int(f.readline())
    for i in range(0,count):
        line = f.readline()
        splits = line.split(',')
        thread = EmbThread()
        thread.catalog_number = splits[0]
        thread.set_color(int(splits[1]), int(splits[2]), int(splits[3]))
        out.add_thread(thread)



from .EmbConstant import *
from .WriteHelper import write_int_8, write_int_16le, write_int_32le

FULL_JUMP = False
ROUND = True
MAX_JUMP_DISTANCE = 124  # 0xFCFF/2 16 bit signed, FD,FE,FF reserved.
MAX_STITCH_DISTANCE = 124


def write_xxx_header_b(pattern, f):
    stitches = pattern.stitches
    for i in range(0, 0x17):
        write_int_8(f, 0x00)
    write_int_32le(f, len(stitches) - 1)
    # END command not called a command.
    for i in range(0, 0x0C):
        write_int_8(f, 0x00)
    write_int_32le(f, len(pattern.threadlist))
    write_int_16le(f, 0x0000)

    extends = pattern.bounds()
    width = int(extends[2] - extends[0])
    height = int(extends[3] - extends[1])
    write_int_16le(f, width)
    write_int_16le(f, height)
    write_int_16le(f, int(stitches[-1][0]))
    write_int_16le(f, int(-stitches[-1][1]))
    write_int_16le(f, int(-extends[0]))
    write_int_16le(f, int(extends[3]))
    for i in range(0, 0x42):
        write_int_8(f, 0x00)
    write_int_16le(f, 0x00)  # unknown
    write_int_16le(f, 0x00)  # unknown
    for i in range(0, 0x73):
        write_int_8(f, 0x00)
    write_int_16le(f, 0x20)
    for i in range(0, 0x08):
        write_int_8(f, 0x00)


def write_xxx_header_a(pattern, f):
    stitches = pattern.stitches
    for i in range(0, 0x17):
        write_int_8(f, 0x00)
    write_int_32le(f, len(stitches) - 1)
    # END command not called a command.
    for i in range(0, 0x0C):
        write_int_8(f, 0x00)
    write_int_32le(f, len(pattern.threadlist))
    write_int_16le(f, 0x0000)

    write_int_16le(f, int(stitches[-1][0]))  # correct
    write_int_16le(f, int(-stitches[-1][1]))  # correct
    for i in range(0, 0x85):
        write_int_8(f, 0x00)
    f.write(b'XXX')
    for i in range(0, 0x39):
        write_int_8(f, 0x00)
    write_int_16le(f, 0x20)
    for i in range(0, 0x08):
        write_int_8(f, 0x00)


def write_xxx_stitches(pattern, f):
    stitches = pattern.stitches
    xx = 0
    yy = 0
    for stitch in stitches:
        x = stitch[0]
        y = stitch[1]
        data = stitch[2] & COMMAND_MASK
        dx = int(round(x - xx))
        dy = int(round(y - yy))
        xx += dx
        yy += dy
        if data == COLOR_CHANGE or data == STOP:
            write_int_8(f, 0x7F)
            write_int_8(f, 0x08)
            write_int_8(f, dx)
            write_int_8(f, -dy)
            continue
        if data == END:
            break
        if data == STITCH:
            if -124 < dx < 124 and -124 < dy < 124:
                write_int_8(f, dx)
                write_int_8(f, -dy)
                continue
            else:
                write_int_8(f, 0x7D)
                write_int_16le(f, dx)
                write_int_16le(f, -dy)
                continue
        if data == TRIM:
            write_int_8(f, 0x7F)
            write_int_8(f, 0x03)
            write_int_8(f, dx)
            write_int_8(f, -dy)
            continue
        if data == JUMP:
            write_int_8(f, 0x7F)
            write_int_8(f, 0x01)
            write_int_8(f, dx)
            write_int_8(f, -dy)
            continue


def write_xxx_colors(pattern, f):
    write_int_8(f, 0x00)
    write_int_8(f, 0x00)
    current_color = 0
    colors = pattern.threadlist
    for color in colors:
        write_int_8(f, 0x00)
        write_int_8(f, color.get_red())
        write_int_8(f, color.get_green())
        write_int_8(f, color.get_blue())
        current_color += 1
    for i in range(0, 21 - current_color):
        write_int_32le(f, 0x00000000)
    write_int_32le(f, 0xFFFFFF00)
    write_int_8(f, 0x00)
    write_int_8(f, 0x01)


def write(pattern, f, settings=None):
    write_xxx_header_b(pattern, f)
    place_holder_for_end_of_stitches = f.tell()
    write_int_32le(f, 0x00000000)
    write_xxx_stitches(pattern, f)
    end_of_stitches = f.tell()
    f.seek(place_holder_for_end_of_stitches, 0)
    write_int_32le(f, end_of_stitches)
    f.seek(end_of_stitches, 0)
    write_int_8(f, 0x7F)
    write_int_8(f, 0x7F)
    write_int_8(f, 0x02)
    write_int_8(f, 0x14)
    write_xxx_colors(pattern, f)



from .EmbThread import EmbThread
from .ReadHelper import signed8, read_int_8, read_int_24be


def read(f, out, settings=None):
    f.seek(0x20E, 0)
    while True:
        if read_int_8(f) == 0x45:
            thread = EmbThread()
            thread.color = read_int_24be(f)
            read_int_8(f)  # Should be 0x20 " "
            out.add_thread(thread)
        else:
            break
    f.seek(0x600, 0)

    count = 0
    while True:
        count += 1
        byte = bytearray(f.read(3))
        if len(byte) != 3:
            break
        x = byte[0]
        y = byte[1]
        ctrl = byte[2]
        if ctrl == 0x80:
            out.stitch(signed8(x), -signed8(y))
            continue
        elif ctrl == 0x81:
            if count > 1:  # This might rather be a needle change.
                out.color_change()
        elif ctrl == 0x90:
            if x == 0 and y == 0:
                out.trim()
            else:
                out.move(signed8(x), -signed8(y))
            continue
        elif ctrl == 0x40:
            out.stop()
            continue
        elif ctrl == 0x86:
            out.trim()
            continue
        elif ctrl == 0x8F:
            break
        else:
            break  # Dunno why it got here.
    out.end()



from .EmbFunctions import *

ENCODE = False
WRITE_FILE_IN_TEXT_MODE = True


def decoded_name(names, data):
    command = decode_embroidery_command(data)
    try:
        name = names[command[0]]
        if command[1] is not None:
            name = name + " t" + str(command[1])
        if command[2] is not None:
            name = name + " n" + str(command[2])
        if command[3] is not None:
            name = name + " o" + str(command[3])
    except (IndexError, KeyError):
        name = "UNKNOWN " + str(data)
    return name


def write(pattern, f, settings=None):
    import json
    names = get_common_name_dictionary()

    metadata = {}
    for item in pattern.extras.items():
        key, value = item
        if isinstance(value, str):
            metadata[key] = value
        elif isinstance(value, int):
            metadata[key] = value
        elif isinstance(value, float):
            metadata[key] = value

    json_normal = {
        "threadlist": [
            {
                "color": thread.hex_color(),
                "description": thread.description,
                "catalog_number": thread.catalog_number,
                "details": thread.details,
                "brand": thread.brand,
                "chart": thread.chart,
                "weight": thread.weight
            }
            for thread in pattern.threadlist
        ],
        "stitches": [[s[0], s[1], str(decoded_name(names, s[2]))] for s in pattern.stitches],
        "extras": metadata
    }
    json.dump(json_normal, f, indent=4)



from __future__ import print_function

import unittest

from pattern_for_tests import *


class TestOverloads(unittest.TestCase):

    def position_equals(self, stitches, j, k):
        self.assertEqual(stitches[j][:1], stitches[k][:1])

    def test_pattern_equal(self):
        shift0 = get_shift_pattern()
        shift1 = get_shift_pattern()
        self.assertIsNot(shift0, shift1)
        self.assertEqual(shift0, shift1)

    def test_pattern_merge(self):
        shift0 = get_shift_pattern()
        shift1 = get_shift_pattern()
        shift0.add_command(END)
        last_pos = len(shift0) - 1
        self.assertEqual(shift0[last_pos][2], END)
        shift0 += shift1
        self.assertNotEqual(shift0[last_pos][2], END)

        shift2 = get_shift_pattern()
        shift3 = get_shift_pattern()
        shift2 += shift3

        self.assertEqual(shift0, shift2)

        shift4 = get_shift_pattern()

        self.assertEqual(shift4, shift3)
        self.assertEqual(shift4, shift1)
        shift1["name"] = "shifty"

        self.assertNotEqual(shift4, shift1)

    def test_pattern_merge_color(self):
        p0 = EmbPattern()
        p0 += "blue"
        p0 += ((0, 0), (1, 1), (2, 2))
        p0 += "red"
        p0 += ((4, 4))

        p1 = EmbPattern()
        p1 += "red"
        p1 += ((0, 0), (1, 1), (2, 2)) * 10
        p1 += "yellow"
        p1 += ((4, 4))

        p0 += p1
        self.assertEqual(p0.count_color_changes(), 2)

    def test_pattern_merge_color2(self):
        p0 = EmbPattern()
        p0 += COLOR_BREAK
        p0 += "blue"
        p0 += ((0, 0), (1, 1), (2, 2))
        p0 += "red"
        p0 += ((4, 4))

        p1 = EmbPattern()
        p1 += COLOR_BREAK
        p1 += "red"
        p1 += ((0, 0), (1, 1), (2, 2))
        p1 += "yellow"
        p1 += ((4, 4))

        p0 += p1
        self.assertEqual(p0.count_color_changes(), 2)

    def test_thread_equal(self):
        t0 = EmbThread("red")
        t1 = EmbThread("#F00")
        self.assertEqual(t0, t1)

    def test_matrix(self):
        m0 = EmbMatrix()
        m1 = EmbMatrix()
        m0.post_scale(2)
        m1.post_rotate(30)
        catted = m0.__matmul__(m1)  # might run in 2.7
        m2 = EmbMatrix()
        m2.post_scale(2)
        m2.post_rotate(30)
        self.assertEqual(catted, m2)





from .EmbThreadSew import get_thread_set
from .ReadHelper import read_int_16le, signed8


def read_sew_stitches(f, out):
    while True:
        b = bytearray(f.read(2))
        if len(b) != 2:
            break
        if b[0] != 0x80:
            out.stitch(signed8(b[0]), -signed8(b[1]))
            continue
        control = b[1]
        b = bytearray(f.read(2))
        if len(b) != 2:
            break
        if control & 1:
            out.color_change()
            continue
        if control == 0x04 or control == 0x02:
            out.move(signed8(b[0]), -signed8(b[1]))
            continue
        if control == 0x10:
            out.stitch(signed8(b[0]), -signed8(b[1]))
            continue
        break
    out.end()


def read(f, out, settings=None):
    threads = get_thread_set()
    colors = read_int_16le(f)
    for c in range(0, colors):
        index = read_int_16le(f)
        index %= len(threads)
        out.add_thread(threads[index])

    f.seek(0x1D78, 0)
    read_sew_stitches(f, out);



def expand(data, uncompressed_size=None):
    emb_compress = EmbCompress()
    return emb_compress.decompress(data, uncompressed_size)


def compress(data):
    size = len(data)
    return bytearray([(size >> 0) & 0xFF, (size >> 8) & 0xFF, 0x02, 0xA0, 0x01, 0xFE]) + data


class Huffman:
    def __init__(self, lengths=None, value=0):
        self.default_value = value
        self.lengths = lengths
        self.table = None
        self.table_width = 0

    def build_table(self):
        """Build an index huffman table based on the lengths. lowest index value wins in a tie."""
        self.table_width = max(self.lengths)
        self.table = []
        size = (1 << self.table_width)
        for bit_length in range(1, self.table_width + 1):
            size /= 2.0
            for len_index in range(0, len(self.lengths)):
                length = self.lengths[len_index]
                if length == bit_length:
                    self.table += [len_index] * int(size)

    def lookup(self, byte_lookup):
        """lookup into the index, returns value and length
        must be requested with 2 bytes."""
        if self.table is None:
            return self.default_value, 0
        v = self.table[byte_lookup >> (16 - self.table_width)]
        return v, self.lengths[v]


class EmbCompress:
    def __init__(self):
        self.bit_position = 0
        self.input_data = None
        self.block_elements = None
        self.character_huffman = None
        self.distance_huffman = None

    def get_bits(self, start_pos_in_bits, length):
        end_pos_in_bits = start_pos_in_bits + length - 1
        start_pos_in_bytes = int(start_pos_in_bits / 8)
        end_pos_in_bytes = int(end_pos_in_bits / 8)
        value = 0
        for i in range(start_pos_in_bytes, end_pos_in_bytes + 1):
            value <<= 8
            try:
                value |= self.input_data[i] & 0xFF
            except IndexError:
                pass
        unused_bits_right_of_sample = (8 - (end_pos_in_bits + 1) % 8) % 8
        mask_sample_bits = (1 << length) - 1
        original = (value >> unused_bits_right_of_sample) & mask_sample_bits
        return original

    def pop(self, bit_count):
        value = self.peek(bit_count)
        self.slide(bit_count)
        return value

    def peek(self, bit_count):
        return self.get_bits(self.bit_position, bit_count)

    def slide(self, bit_count):
        self.bit_position += bit_count

    def read_variable_length(self):
        m = self.pop(3)
        if m != 7:
            return m
        for q in range(0, 13):  # max read is 16 bit, 3 bits already used. It can't exceed 16-3
            s = self.pop(1)
            if s == 1:
                m += 1
            else:
                break
        return m

    def load_character_length_huffman(self):
        count = self.pop(5)
        if count == 0:
            v = self.pop(5)
            huffman = Huffman(value=v)
        else:
            huffman_code_lengths = [0] * count
            index = 0
            while index < count:
                if index == 3:  # Special index 3, skip up to 3 elements.
                    index += self.pop(2)
                huffman_code_lengths[index] = self.read_variable_length()
                index += 1
            huffman = Huffman(huffman_code_lengths, 8)
            huffman.build_table()
        return huffman

    def load_character_huffman(self, length_huffman):
        count = self.pop(9)
        if count == 0:
            v = self.pop(9)
            huffman = Huffman(value=v)
        else:
            huffman_code_lengths = [0] * count
            index = 0
            while index < count:
                h = length_huffman.lookup(self.peek(16))
                c = h[0]
                self.slide(h[1])
                if c == 0:  # C == 0, skip 1.
                    c = 1
                    index += c
                elif c == 1:  # C == 1, skip 3 + read(4)
                    c = 3 + self.pop(4)
                    index += c
                elif c == 2:  # C == 2, skip 20 + read(9)
                    c = 20 + self.pop(9)
                    index += c
                else:
                    c -= 2
                    huffman_code_lengths[index] = c
                    index += 1
            huffman = Huffman(huffman_code_lengths)
            huffman.build_table()
        return huffman

    def load_distance_huffman(self):
        count = self.pop(5)
        if count == 0:
            v = self.pop(5)
            huffman = Huffman(value=v)
        else:
            index = 0
            lengths = [0] * count
            for i in range(0, count):
                lengths[index] = self.read_variable_length()
                index += 1
            huffman = Huffman(lengths)
            huffman.build_table()
        return huffman

    def load_block(self):
        self.block_elements = self.pop(16)
        character_length_huffman = self.load_character_length_huffman()
        self.character_huffman = self.load_character_huffman(character_length_huffman)
        self.distance_huffman = self.load_distance_huffman()

    def get_token(self):
        if self.block_elements <= 0:
            self.load_block()
        self.block_elements -= 1
        h = self.character_huffman.lookup(self.peek(16))
        self.slide(h[1])
        return h[0]

    def get_position(self):
        h = self.distance_huffman.lookup(self.peek(16))
        self.slide(h[1])
        if h[0] == 0:
            return 0
        v = h[0] - 1
        v = (1 << v) + self.pop(v)
        return v

    def decompress(self, input_data, uncompressed_size=None):
        self.input_data = input_data
        output_data = []
        self.block_elements = -1
        bits_total = (len(input_data) * 8)
        while bits_total > self.bit_position and (uncompressed_size is None or len(output_data) <= uncompressed_size):
            character = self.get_token()
            if character <= 255:  # literal.
                output_data.append(character)
            elif character == 510:
                break  # END
            else:
                length = character - 253  # Min length is 3. 256-253=3.
                back = self.get_position() + 1
                position = len(output_data) - back
                if back > length:
                    # Entire lookback is already within output data.
                    output_data += output_data[position:position + length]
                else:
                    # Will read & write the same data at some point.
                    for i in range(position, position + length):
                        output_data.append(output_data[i])
        return output_data



from .EmbThread import EmbThread


def get_thread_set():
    return [
        None,  #EmbThreadJef(0x000000, "Placeholder", "000"),
        EmbThreadJef(0x000000, "Black", "002"),
        EmbThreadJef(0xffffff, "White", "001"),
        EmbThreadJef(0xffff17, "Yellow", "204"),
        EmbThreadJef(0xff6600, "Orange", "203"),
        EmbThreadJef(0x2f5933, "Olive Green", "219"),
        EmbThreadJef(0x237336, "Green", "226"),
        EmbThreadJef(0x65c2c8, "Sky", "217"),
        EmbThreadJef(0xab5a96, "Purple", "208"),
        EmbThreadJef(0xf669a0, "Pink", "201"),
        EmbThreadJef(0xff0000, "Red", "225"),
        EmbThreadJef(0xb1704e, "Brown", "214"),
        EmbThreadJef(0x0b2f84, "Blue", "207"),
        EmbThreadJef(0xe4c35d, "Gold", "003"),
        EmbThreadJef(0x481a05, "Dark Brown", "205"),
        EmbThreadJef(0xac9cc7, "Pale Violet", "209"),
        EmbThreadJef(0xfcf294, "Pale Yellow", "210"),
        EmbThreadJef(0xf999b7, "Pale Pink", "211"),
        EmbThreadJef(0xfab381, "Peach", "212"),
        EmbThreadJef(0xc9a480, "Beige", "213"),
        EmbThreadJef(0x970533, "Wine Red", "215"),
        EmbThreadJef(0xa0b8cc, "Pale Sky", "216"),
        EmbThreadJef(0x7fc21c, "Yellow Green", "218"),
        EmbThreadJef(0xe5e5e5, "Silver Gray", "220"),
        EmbThreadJef(0x889b9b, "Gray", "221"),
        EmbThreadJef(0x98d6bd, "Pale Aqua", "227"),
        EmbThreadJef(0xb2e1e3, "Baby Blue", "228"),
        EmbThreadJef(0x368ba0, "Powder Blue", "229"),
        EmbThreadJef(0x4f83ab, "Bright Blue", "230"),
        EmbThreadJef(0x386a91, "Slate Blue", "231"),
        EmbThreadJef(0x071650, "Navy Blue", "232"),
        EmbThreadJef(0xf999a2, "Salmon Pink", "233"),
        EmbThreadJef(0xf9676b, "Coral", "234"),
        EmbThreadJef(0xe3311f, "Burnt Orange", "235"),
        EmbThreadJef(0xe2a188, "Cinnamon", "236"),
        EmbThreadJef(0xb59474, "Umber", "237"),
        EmbThreadJef(0xe4cf99, "Blond", "238"),
        EmbThreadJef(0xffcb00, "Sunflower", "239"),
        EmbThreadJef(0xe1add4, "Orchid Pink", "240"),
        EmbThreadJef(0xc3007e, "Peony Purple", "241"),
        EmbThreadJef(0x80004b, "Burgundy", "242"),
        EmbThreadJef(0x540571, "Royal Purple", "243"),
        EmbThreadJef(0xb10525, "Cardinal Red", "244"),
        EmbThreadJef(0xcae0c0, "Opal Green", "245"),
        EmbThreadJef(0x899856, "Moss Green", "246"),
        EmbThreadJef(0x5c941a, "Meadow Green", "247"),
        EmbThreadJef(0x003114, "Dark Green", "248"),
        EmbThreadJef(0x5dae94, "Aquamarine", "249"),
        EmbThreadJef(0x4cbf8f, "Emerald Green", "250"),
        EmbThreadJef(0x007772, "Peacock Green", "251"),
        EmbThreadJef(0x595b61, "Dark Gray", "252"),
        EmbThreadJef(0xfffff2, "Ivory White", "253"),
        EmbThreadJef(0xb15818, "Hazel", "254"),
        EmbThreadJef(0xcb8a07, "Toast", "255"),
        EmbThreadJef(0x986c80, "Salmon", "256"),
        EmbThreadJef(0x98692d, "Cocoa Brown", "257"),
        EmbThreadJef(0x4d3419, "Sienna", "258"),
        EmbThreadJef(0x4c330b, "Sepia", "259"),
        EmbThreadJef(0x33200a, "Dark Sepia", "260"),
        EmbThreadJef(0x523a97, "Violet Blue", "261"),
        EmbThreadJef(0x0d217e, "Blue Ink", "262"),
        EmbThreadJef(0x1e77ac, "Sola Blue", "263"),
        EmbThreadJef(0xb2dd53, "Green Dust", "264"),
        EmbThreadJef(0xf33689, "Crimson", "265"),
        EmbThreadJef(0xde649e, "Floral Pink", "266"),
        EmbThreadJef(0x984161, "Wine", "267"),
        EmbThreadJef(0x4c5612, "Olive Drab", "268"),
        EmbThreadJef(0x4c881f, "Meadow", "269"),
        EmbThreadJef(0xe4de79, "Mustard", "270"),
        EmbThreadJef(0xcb8a1a, "Yellow Ocher", "271"),
        EmbThreadJef(0xcba21c, "Old Gold", "272"),
        EmbThreadJef(0xff9805, "Honey Dew", "273"),
        EmbThreadJef(0xfcb257, "Tangerine", "274"),
        EmbThreadJef(0xffe505, "Canary Yellow", "275"),
        EmbThreadJef(0xf0331f, "Vermilion", "202"),
        EmbThreadJef(0x1a842d, "Bright Green", "206"),
        EmbThreadJef(0x386cae, "Ocean Blue", "222"),
        EmbThreadJef(0xe3c4b4, "Beige Gray", "223"),
        EmbThreadJef(0xe3ac81, "Bamboo", "224")
    ]


class EmbThreadJef(EmbThread):
    def __init__(self, color, description, catalog_number):
        EmbThread.__init__(self)
        self.color = color
        self.description = description
        self.catalog_number = catalog_number
        self.brand = "Jef"
        self.chart = "Jef"



from .WriteHelper import write_int_8, write_int_32be, write_int_16be, write_string_utf8

ENCODE = False


def patch_byte_offset(stream, offset):
    current_pos = stream.tell()
    stream.seek(offset, 0)  # Absolute position seek.
    position = current_pos - offset - 4  # 4 bytes int32
    write_int_32be(stream, position)
    stream.seek(current_pos, 0)  # Absolute position seek.


def write(pattern, f, settings=None):
    write_int_32be(f, 1)
    write_int_32be(f, 8)
    placeholder = f.tell()
    write_int_32be(f, 0)  # Placeholder.
    write_int_32be(f, len(pattern.threadlist))
    index = 0
    for thread in pattern.threadlist:
        details = thread.description
        if details is None:
            details = "Unknown"
        chart = thread.chart
        if chart is None:
            chart = "Unknown"
        write_int_16be(f, 11 + len(details) + len(chart))  # 2 + 2 + 1 + 1 + 1 + 2 + d + 1 + c + 1 = 11 + d + c
        write_int_16be(f, index)  # record index
        index += 1
        write_int_8(f, thread.get_red())
        write_int_8(f, thread.get_green())
        write_int_8(f, thread.get_blue())
        write_int_16be(f, index)  # needle number
        write_string_utf8(f, details)
        write_int_8(f, 0)
        write_string_utf8(f, chart)
        write_int_8(f, 0)
    patch_byte_offset(f, placeholder)



from .WriteHelper import write_int_8

ENCODE = False


def write(pattern, f, settings=None):
    if len(pattern.threadlist) > 0:
        for thread in pattern.threadlist:
            write_int_8(f, thread.get_red())
            write_int_8(f, thread.get_green())
            write_int_8(f, thread.get_blue())
            write_int_8(f, 0)



def read_ksm_stitches(f, out):
    trimmed = False
    stitched_yet = False
    count = 0
    while True:
        count += 1
        byte = bytearray(f.read(3))
        if len(byte) != 3:
            break

        y = -byte[0]
        x = byte[1]
        ctrl = byte[2]

        if ctrl & 0x40 != 0:
            x = -x
        if ctrl & 0x20 != 0:
            y = -y

        ctrl &= 0b11111
        if x != 0 or y != 0:
            if trimmed:  # any x, y gets executed regardless.
                out.move(x, y)
            else:
                out.stitch(x, y)
                stitched_yet = True

        if ctrl == 0x00:
            continue
        # print(str(f), " ", str(count), " ", str("{0:b}").format(ctrl), " 0x%0.2X " % ctrl, x, " ", y)

        if ctrl == 0x07 or ctrl == 0x13 or ctrl == 0x1D:
            if stitched_yet:
                out.trim()
            trimmed = True
            continue
        if 0x17 <= ctrl <= 0x19:  # start sewing again.
            trimmed = False
            continue
        if 0x0B <= ctrl <= 0x12:
            needle = ctrl - 0x0A
            out.needle_change(needle)
            trimmed = True
            continue
        if ctrl == 0x05:
            out.stop()
            continue
        if ctrl == 0x1B:  # Called before end command
            trimmed = False
            continue
        if ctrl == 0x08:  # End command #88 zero direction.
            break
        break  # Uncaught Control
    out.end()


def read(f, out, settings=None):
    f.seek(0x200, 0)
    read_ksm_stitches(f, out)



from .ReadHelper import read_int_8, read_int_24be, signed24, read_int_16be

PC_SIZE_CONVERSION_RATIO = 5.0 / 3.0


def read_pc_file(f, out, settings=None):
    pcm_threads = [
        {"color": 0x000000, "description": "PCM Color 1"},
        {"color": 0x000080, "description": "PCM Color 2"},
        {"color": 0x0000FF, "description": "PCM Color 3"},
        {"color": 0x008080, "description": "PCM Color 4"},
        {"color": 0x00FFFF, "description": "PCM Color 5"},
        {"color": 0x800080, "description": "PCM Color 6"},
        {"color": 0xFF00FF, "description": "PCM Color 7"},
        {"color": 0x800000, "description": "PCM Color 8"},
        {"color": 0xFF0000, "description": "PCM Color 9"},
        {"color": 0x008000, "description": "PCM Color 10"},
        {"color": 0x00FF00, "description": "PCM Color 11"},
        {"color": 0x808000, "description": "PCM Color 12"},
        {"color": 0xFFFF00, "description": "PCM Color 13"},
        {"color": 0x808080, "description": "PCM Color 14"},
        {"color": 0xC0C0C0, "description": "PCM Color 15"},
        {"color": 0xFFFFFF, "description": "PCM Color 16"},
    ]

    f.seek(2, 0)

    colors = read_int_16be(f)
    if colors is None:
        return  # File is blank.
    for i in range(0, colors):
        color_index = read_int_16be(f)
        thread = pcm_threads[color_index]
        out.add_thread(thread)

    stitch_count = read_int_16be(f)
    while True:
        x = read_int_24be(f)
        c0 = read_int_8(f)
        y = read_int_24be(f)
        c1 = read_int_8(f)
        ctrl = read_int_8(f)
        if ctrl is None:
            break
        x = signed24(x)
        y = -signed24(y)
        x *= PC_SIZE_CONVERSION_RATIO
        y *= PC_SIZE_CONVERSION_RATIO
        if ctrl == 0x00:
            out.stitch_abs(x, y)
            continue
        if ctrl & 0x01:
            out.color_change()
            continue
        if ctrl & 0x04:
            out.move_abs(x, y)
            continue
        break  # Uncaught Control
    out.end()


def read(f, out, settings=None):
    read_pc_file(f, out)



def read_inb_stitches(f, out):
    count = 0
    while True:
        count += 1
        byte = bytearray(f.read(3))
        if len(byte) != 3:
            break
        x = byte[0]
        y = -byte[1]
        ctrl = byte[2]
        if ctrl & 0x20 != 0:
            y = -y
        if ctrl & 0x40 != 0:
            x = -x
        if (ctrl & 0b1111) == 0x00:
            out.stitch(x, y)
            continue
        if (ctrl & 0b1111) == 0x01:
            out.color_change(x, y)
            continue
        if (ctrl & 0b1111) == 0x02:
            out.move(x, y)
            continue
        if ctrl == 0x04:
            break
        break  # Uncaught Control
    out.end()


def read(f, out, settings=None):
    f.seek(0x2000, 0)
    read_inb_stitches(f, out)



import struct


def write_int_array_8(stream, int_array):
    for value in int_array:
        v = bytes(bytearray([
            value & 0xFF,
        ]))
        stream.write(v)


def write_int_8(stream, value):
    v = bytes(bytearray([
        value & 0xFF,
    ]))
    stream.write(v)


def write_int_16le(stream, value):
    v = bytes(bytearray([
        (value >> 0) & 0xFF,
        (value >> 8) & 0xFF,
    ]))
    stream.write(v)


def write_int_16be(stream, value):
    v = bytes(bytearray([
        (value >> 8) & 0xFF,
        (value >> 0) & 0xFF,
    ]))
    stream.write(v)


def write_int_24le(stream, value):
    v = bytes(bytearray([
        (value >> 0) & 0xFF,
        (value >> 8) & 0xFF,
        (value >> 16) & 0xFF,
    ]))
    stream.write(v)


def write_int_24be(stream, value):
    v = bytes(bytearray([
        (value >> 16) & 0xFF,
        (value >> 8) & 0xFF,
        (value >> 0) & 0xFF,
    ]))
    stream.write(v)


def write_int_32le(stream, value):
    v = bytes(bytearray([
        (value >> 0) & 0xFF,
        (value >> 8) & 0xFF,
        (value >> 16) & 0xFF,
        (value >> 24) & 0xFF
    ]))
    stream.write(v)


def write_int_32be(stream, value):
    v = bytes(bytearray([
        (value >> 24) & 0xFF,
        (value >> 16) & 0xFF,
        (value >> 8) & 0xFF,
        (value >> 0) & 0xFF
    ]))
    stream.write(v)


def write_float_32le(stream, value):
    stream.write(struct.pack("<f", float(value)))


def write_string(stream, string, encoding='utf8'):
    # python 2,3 code
    try:
        stream.write(bytes(string).encode(encoding))
    except TypeError:
        stream.write(bytes(string, encoding))


def write_string_utf8(stream, string):
    # python 2,3 code
    try:
        stream.write(bytes(string).encode('utf8'))
    except TypeError:
        stream.write(bytes(string, 'utf8'))



from .ReadHelper import signed8, read_int_32le


def read_zhs_stitches(f, out):
    count = 0
    while True:
        count += 1
        b = bytearray(f.read(3))
        if len(b) != 3:
            break
        ctrl = b[0]
        if ctrl == 0x10:
            pass
        x = signed8(b[1])
        y = signed8(b[2])
        if ctrl == 0x02:
            out.stitch(x, y)
            continue
        if ctrl == 0x01:
            out.move(x, y)
            continue
        if ctrl == 0x04:
            out.color_change()
            continue
        if ctrl == 0x80:
            break
    out.end()


def read(f, out, settings=None):
    f.seek(0x0F, 0)
    stitch_start_position = read_int_32le(f)
    f.seek(stitch_start_position, 0)
    read_zhs_stitches(f, out)
